"""
用来读取本地的股票数据
"""
import pandas as pd
import numpy as np
import yangke.stock.globalVar as gv
import torch.utils.data as tud
import os
# 类似于 pymysql中的游标
from yangke.dataset.mysql import has_table, \
    read_dataframe_from_mysql


class StockDataset(tud.Dataset):
    """
    构建pytorch的dataset
    """

    def __init__(self, stock_codes: list = None):
        """
        如果stock_codes为一个str，则生成该股票代码的数据集
        如果stock_codes为一个list，则生成该list中所有股票的数据集
        如果stock_codes为None，则默认生成所有有效股票的数据集

        :param stock_codes:
        """
        super(StockDataset, self).__init__()
        if stock_codes is None:
            stock_codes = gv.tsd.download_all_stocks()
        if isinstance(stock_codes, list):
            # 默认选择全部股票数据，如果传入stock_codes，则只加载该列表中的股票
            self.stock_codes = gv.tsd.download_all_stocks(stock_codes)  # 该方法会过滤数据不正常的股票
            # self.files = [os.path.join(gv.dataDirectory, code + ".csv") for code in self.stock_codes]
            current_total_length = 0
            self.index2code = []
            for code in stock_codes:
                result = get_data_days_list(code)
                if result is not None:
                    x, _ = result
                    self.index2code.append(
                        {'range': range[current_total_length, current_total_length + len(x)], 'code': code})
                    current_total_length = current_total_length + len(x)

            self.length = current_total_length
        else:
            self.stock_codes = str(stock_codes)
            result = get_data_days_list(self.stock_codes)
            if result is not None:
                x, y = result
            self.length = len(x)

    def __getitem__(self, index):
        if isinstance(self.stock_codes, list):
            code, pos = self.get_code_pos(index)
            result = get_data_days_list(code)
            x, y = result
        else:
            result = get_data_days_list(self.stock_codes)
            if result is not None:
                x, y = result
        return x[pos], y[pos]

    def __len__(self):
        # 返回整个数据集的长度
        return self.length

    def __repr__(self):
        if isinstance(self.stock_codes, list):
            return "包含以下股票的数据：{}".format(self.stock_codes)
        else:
            return "包含股票{}的数据".format(self.stock_codes)

    def get_code_pos(self, index):
        for i in range(len(self.index2code)):
            local_range = self.index2code[i]['range']
            bottom_index = next(iter(local_range))
            if index in local_range:
                pos = index - bottom_index
                code = self.index2code[i]['code']
                break
        return code, pos


def get_data(stock_code, col_x=None, col_y=None, normalization=None) -> np.ndarray:
    """
    读取股票本地数据文件中指定columns的数据，返回ndarray格式，指定归一化方法可以进行归一化

    :param stock_code: 股票代码
    :param col_x: 需要读入的数据列
    :param col_y: 股票涨幅，如果传入的是列表，会把涨幅按列表分类，例[1, 5]，列表中的值应大于-10，小于10
    :param normalization: 是否需要归一化，默认不归一化处理，需要归一化处理时传入归一化方法，可取值'min-max'、'z-score'
    :return: x_values, y_values，分别表示输入数据和对应的标签值

    example:

        get_data('C:/stock/600006.csv', None, col_y=[1, 5])

        col_y=[1, 5]，
        则获得的标签中0代表股票涨幅小于1%，1代表涨幅位于[1, 5)，2代表涨幅处于[5,)

        col_y=[1]，则获得的标签中0代表股票涨幅小于1%，1表示股票涨幅大于等于1%。
    """
    gv.tsd.download(stock_code)  # 下载股票数据到本地，如果存在本地数据，则该语句相当于什么也不做
    # os.path.join(gv.dataDirectory, '{}.csv'.format(stock_code))
    # data = pd.read_csv(stock_code, encoding=gv.encoding)  # 相当于要让这句话常驻内存
    data = read_data_from_mysql_csv(stock_code, encoding=gv.encoding)  # 更新读取csv文件的方法，利用redis将数据驻留内存，多次调用加快速度
    if col_x is None:
        col_x = ['close']

    x = data[col_x].copy()  # 这里不加.copy()则报警告，因为修改x,y会修改data中对应的列值，pandas认为这可能是有问题的
    y = data['p_change'].copy()  # 股票涨幅位于区间[-10, 10]内；
    # =================================如果神经网络的输出是类别，这里把连续的y值按区间分类================================
    if isinstance(col_y, list):
        if len(col_y) == 0:
            gv.logger.warning("传入的col_y是一个空列表！")
        for i in col_y:
            if i < -10 or i > 10:
                gv.logger.warning("col_y的值超范围，涨跌幅限制在-0.1~0.1之间，传入的列表值达到{}".format(i))
                break
        # 按区间将标签分类，这里标签修改的顺序需要注意，
        # 修改后的标签的值不应落在区间[-10,10]内，所以标签0放在最后一个修改。
        y.loc[y > col_y[-1]] = len(col_y) * 100
        for i in range(len(col_y) - 1, 0, -1):
            condition_series = y[y > col_y[i - 1]][y < col_y[i]]  # 符合条件的行
            y[condition_series.index] = i * 100  # condition_series对应的行号的值设为i * 100
        y.loc[y < col_y[0]] = 0
        y = (y / 100).astype(int)
    # =================================如果神经网络的输出是类别，这里把连续的y值按区间分类================================
    if normalization == 'min-max':
        x = (x - x.min()) / (x.max() - x.min())  # df.min()默认是按行操作，即axis=0，将df形状的第一个数压缩为1
    elif normalization == 'z-score':
        x = (x - x.mean()) / x.std_range()  # 据说，pandas的std函数和numpy的std函数有差别，但是不影响神经网络训练结果

    return x.values, y.values  # 将DataFrame或Series转为ndarray


def get_data_stocks(stock_codes: list = None, col_x=None, col_y=None, normalization=None):
    """
    获取所有股票的数据

    从纯数据预测的角度来讲，所有股票的股价变化趋势都应是一样的

    :param stock_codes:
    :param col_x:
    :param col_y:
    :param normalization:
    :return:
    """
    stock_codes = gv.tsd.download_all_stocks(stock_codes)

    for stock_code in stock_codes:
        x, y = get_data(stock_code, col_x=None, col_y=None, normalization=None)


def get_data_days_list(stock_code, days=30, col_x=None, col_y=None, normalization=None, split=None,
                       need_pred=False):
    """
    构建神经网络训练数据集

    :param stock_code: 文件
    :param days: 根据过去days天的股票数据，预测接下来一天的涨幅
    :param col_x: x中的数据列
    :param col_y: 股票涨幅，如果传入的是列表，会把涨幅按列表分类，用于神经网络分类预测，例[1, 5]，列表中的值应大于-10，小于10
    :param normalization: 是否需要归一化，默认不归一化处理，需要归一化处理时传入归一化方法，可取值'min-max'、'z-score'
    :param split: split是None则不分割，split是(0,1)的小数，则分为两部分，如果split是长度为2的列表，如[0.7, 0.2]，则分为三部分
    :param need_pred: 是否返回需要预测的x数据
    :return: tuple(train_x, train_y, valid_x, valid_y, [...], pred_x)，数据类型是np.array
    """
    result_x, result_y = [], []
    x, y = get_data(stock_code, col_x=col_x, col_y=col_y, normalization=normalization)
    data_num = len(y)
    if data_num <= days + 100:
        gv.logger.error("文件{}中数据太少，不足以生成神经网络数据！data_num={}".format(stock_code, data_num))
        return None
    for i in range(data_num - days):  # 至少要有31天的数据，才能产生一个神经网络数据，即根据过去30天数据预测下一天的数据
        temp_x = []
        for j in range(days):
            temp_x.append(x[i + j])
        result_x.append(temp_x)
        result_y.append(y[i + days])
    # np.ndarray(x) 是以x为形状参数构建ndarray数组
    # np.array(x) 是以x为数据内容构建ndarray数组
    result = __split_dataset(np.array(result_x), np.array(result_y), split)

    if need_pred:
        pred_x = x[-days:]
        result = list(result)
        result.append(pred_x)
        result = tuple(result)
    return result


def __split_dataset(x, y, split=None):
    """
    将数据按split分为训练集、测试集和验证集。
    split是None则不分割，split是(0,1)之间的小数，则分为两部分，如果split是长度为2的列表，如[0.7, 0.2]，则分为三部分

    :param x: 数据集x，传入的是列表，则返回的tuple中对应的类型也是列表，传入的是np.array，则返回的tuple中对应的类型也是np.array
    :param y: 数据集y
    :param split: 分割数据集的大小
    :return: (x_train, y_train, x_test, y_test, x_valid, y_valid)
    """
    if split is None:
        result = (x, y)
    if isinstance(split, float):
        split = int(len(y) * split)
        x_train, y_train = x[:split], y[:split]
        x_test, y_test = x[split:], y[split:]
        result = (x_train, y_train, x_test, y_test)
    elif isinstance(split, list) and isinstance(split[0], float) and isinstance(split[1], float):
        split0 = int(len(y) * split[0])
        split1 = int(len(y) * (split[0] + split[1]))
        x_train, y_train = x[:split0], y[:split0]
        x_test, y_test = x[split0:split1], y[split0:split1]
        x_valid, y_valid = x[split1:], y[split1:]
        result = (x_train, y_train, x_test, y_test, x_valid, y_valid)
    return result


def get_data_days_generator(stock_code, days=30, col_x=None, col_y=None, split=0.7, normalization=None,
                            need_pred=False):
    """
    作用同get_data_days_list()，只是返回的是generator，目前有问题

    :param stock_code: 文件
    :param days: 根据过去days天的股票数据，预测接下来一天的涨幅
    :param col_x: x中的数据列
    :param col_y: 股票涨幅，如果传入的是列表，会把涨幅按列表分类，用于神经网络分类预测，例[1, 5]，列表中的值应大于-10，小于10
    :param split: split是None则不分割，split是(0,1)的小数，则分为两部分，如果split是长度为2的列表，如[0.7, 0.2]，则分为三部分
    :param normalization: 是否需要归一化，默认不归一化处理，需要归一化处理时传入归一化方法，可取值'min-max'、'z-score'
    :param need_pred: 是否返回需要预测的x数据
    """
    x, y = get_data_days_list(stock_code, days, col_x, col_y, normalization)

    result = __split_dataset(x, y, split)
    x_train, y_train, x_test, y_test = result

    def generator():
        for i in range(len(y)):
            yield x[i], y[i]

    def generator_train():
        for i in range(len(y_train)):
            yield x_train[i], y_train[i]

    def generator_test():
        for i in range(len(y_test)):
            yield x_test[i], y_test[i]

    if need_pred:  # 是否需要返回最后一个需要预测的x数据
        pred_x = x[-days:]
    return generator, generator_train, generator_test, pred_x


def read_data_from_mysql_csv(stock_code, encoding=gv.encoding):
    """
    读取股票基本数据，从mysql数据库或本地读取股票数据

    :param stock_code:
    :param encoding:
    :return:
    """

    # 首先检测mysql服务是否在运行
    # encoding: utf-8

    data = None
    if gv.storage == "mysql":  # 尝试从mysql数据库读取
        if has_table(table_name=f"basic{stock_code}"):
            data = read_dataframe_from_mysql(f"basic{stock_code}")
            if data is not None:
                return data
    else:
        # 如果本地没有mysql服务，就使用简单的csv文件读取
        file = os.path.join(gv.dataDirectory, '{}.csv'.format(stock_code))
        data = pd.read_csv(file, encoding=gv.encoding)
        return data



# game frame
import os.path
import random
import sys
import time

from yangke.common.win.win_x64 import (find_window, capture_pic_undecorated, post_key, do_click,
                                       get_size_of_window, find_pic)
from yangke.objDetect.ocr import ocr
from yangke.base import show_pic, get_pic_size, is_number, draw_element_on_pic, get_draw_element_on_pic_handle
from yangke.common.config import logger
from yangke.common.win.dminput import DM, DMRemote
import cv2


class RectRegion:
    def __init__(self, center=None, left=None, right=None, top=None, bottom=None):
        """
        长方形区域，由上下左右四条边限定的长方形区域，区域位置为区域的中心点
        """
        self.center = center
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom

    def set_xyxy(self, x1, y1, x2, y2):
        self.left = x1
        self.right = x2
        self.top = y1
        self.bottom = y2
        return self

    def x1(self):
        return self.left

    def y1(self):
        return self.top

    def x2(self):
        return self.right

    def y2(self):
        return self.bottom

    def transfer_to_positive(self, width, height):
        """
        将负数表示的坐标转换成正数，负数表示点距离最大值的距离
        """
        self.right = self.right + width if self.right < 0 else self.right
        self.left = self.left + width if self.left < 0 else self.left
        self.top = self.top + height if self.top < 0 else self.top
        self.bottom = self.bottom + height if self.bottom < 0 else self.bottom
        return self.left, self.top, self.right, self.bottom

    def w(self):
        assert self.right > 0 and self.left > 0, "坐标值必须为正才可使用w()方法获取区域宽度"
        return self.right - self.left

    def h(self):
        assert self.right > 0 and self.left > 0, "坐标值必须为正才可使用h()方法获取区域高度"
        return self.bottom - self.top

    def set_xywh(self, x1, y1, w, h):
        self.left = x1
        self.top = y1
        self.right = self.left + w
        self.bottom = self.top + h
        return self

    def get_xyxy(self):
        return self.left, self.top, self.right, self.bottom

    def get_xywh(self):
        return self.left, self.top, self.right - self.left, self.bottom - self.top

    def get_center(self):
        return int((self.left + self.right) / 2), int((self.top + self.bottom) / 2)

    def get_sub_pos(self, start, end, length, axis=0):
        """
        获取Position的子位置

        :param start: 子位置开始的索引
        :param end: 子位置结束的索引
        :param length: 当前位置的长度
        :param axis: 获取子位置的方向
        """
        if axis == 0:
            sep = (self.right - self.left) / length
            return RectRegion().set_xyxy(self.left + sep * start, self.top, self.left + sep * end, self.bottom)


class Map:
    def __init__(self):
        ...


class Chat:
    def __init__(self, region, channels=None):
        """
        聊天面板
        :param region: 聊天面板的区域
        :param channels: 聊天频道列表
        """
        self.region = region
        self.channels = channels


class TimePanel:
    def __init__(self, region, frame: "Frame"):
        self.region = region
        self.frame = frame

    def get_time(self):
        return self.frame.get_text(self.region)


class Task:
    def __init__(self, region_xyxy=None, frame: "Frame" = None):
        self.region_xyxy = region_xyxy  # (x1, y1, x2, y2)
        self.frame: "Frame" | None = frame

    def get_task_text(self, frame: "Frame" = None):
        self.frame = frame or self.frame
        snapshot = self.frame.capture_window_xyxy(*self.region_xyxy)
        text = ocr(snapshot, paragraph=True)
        return text

    def get_text_link_pos_global(self, text: str):
        """
        获取任务面板上指定文字在整个画面上的坐标位置
        """
        snapshot = self.frame.capture_window_xyxy(*self.region_xyxy)
        texts = ocr(snapshot, paragraph=False)
        pos_rel = RectRegion()
        for res in texts:
            p_left_top, p_right_top, p_right_bottom, p_left_bottom = res[0]
            text_, probability = res[1]
            if text in text_:
                pos = RectRegion().set_xyxy(p_left_top[0], p_left_top[1], p_right_bottom[0], p_right_bottom[1])
                start = text_.index(text)
                end = start + len(text)
                pos_rel = pos.get_sub_pos(start, end, len(text_))
                break

        pos_global = self.to_global(pos_rel)
        return pos_global

    def set_frame(self, frame: "Frame"):
        self.frame = frame or self.frame

    def to_global(self, pos_rel: RectRegion | None = None):
        x1, y1, x2, y2 = pos_rel.left, pos_rel.top, pos_rel.right, pos_rel.bottom
        x_offset = self.region_xyxy[0]
        y_offset = self.region_xyxy[1]
        pos_global = RectRegion().set_xyxy(x1 + x_offset, y1 + y_offset, x2 + x_offset, y2 + y_offset)
        return pos_global


class AnchorRegion:
    def __init__(self, x1, y1, x2, y2):
        """
        根据锚点确定的区域类
        示例：
        AnchorRegion("anchor.x1+100", "anchor.y1-10", 200, 400)
        """
        self.x1, self.y1, self.x2, self.y2 = x1, y1, x2, y2
        self.anchor_region: list | tuple | None = None  # 如果是list，表示有多个anchor对象，如果是tuple，表示是一个anchor对象的四个坐标xyxy

    def get_region(self, anchor_region, window_size):
        """
        根据参考点坐标及窗口大小获取AnchorRegion的区域
        可以有多个参考点坐标，此时anchor_region为列表[(x1, y1, x2, y2), (x3, y3, x4, y4)]


        """
        self.anchor_region = anchor_region
        if isinstance(self.x1, str):
            self.x1 = int(self.parse_pos(self.x1))
        if isinstance(self.x2, str):
            self.x2 = int(self.parse_pos(self.x2))
        if isinstance(self.y1, str):
            self.y1 = int(self.parse_pos(self.y1))
        if isinstance(self.y2, str):
            self.y2 = int(self.parse_pos(self.y2))

        if self.x1 < 0:
            self.x1 = self.x1 + window_size[0]
        if self.x2 < 0:
            self.x2 = self.x2 + window_size[0]
        if self.y1 < 0:
            self.y1 = self.y1 + window_size[1]
        if self.y2 < 0:
            self.y2 = self.y2 + window_size[1]
        return self.x1, self.y1, self.x2, self.y2

    def parse_pos(self, pos, anchor=None):
        anchor = anchor or self.anchor_region  # 临时使用的anchor，不能更改self.anchor_region
        if isinstance(anchor, RectRegion):
            anchor = (anchor.left, anchor.top,
                      anchor.right, anchor.bottom)
        if isinstance(anchor, list):
            logger.warning("暂不支持多个anchor")
        elif isinstance(anchor, tuple):
            if "+" in pos:
                _1, _2 = pos.split("+")
                if is_number(_1):
                    _1 = float(_1)
                else:
                    _1 = self.parse_pos(_1, anchor)
                if is_number(_2):
                    _2 = float(_2)
                else:
                    _2 = self.parse_pos(_2, anchor)
                return _1 + _2
            elif "-" in pos:
                _1, _2 = pos.split("-")
                if is_number(_1):
                    _1 = float(_1)
                else:
                    _1 = self.parse_pos(_1, anchor)
                if is_number(_2):
                    _2 = float(_2)
                else:
                    _2 = self.parse_pos(_2, anchor)
                return _1 - _2
            else:
                if pos.endswith("x1"):
                    return anchor[0]
                elif pos.endswith("y1"):
                    return anchor[1]
                elif pos.endswith("x2"):
                    return anchor[2]
                elif pos.endswith("y1"):
                    return anchor[3]


class Offset:
    def __init__(self, x_offset, y_offset=0):
        """
        ALign对象记录了一个元素相对另一个元素的偏移量或其他对齐信息，如果Step类的第二个参数为Offset类型，则表示相对于上一个步骤pos成员变量的偏移值
        """
        self.dx = int(x_offset)
        self.dy = int(y_offset)
        self.x = None
        self.y = None

    def set_anchor_pos(self, rx, ry):
        """
        rx, ry是Align对象参考的其他元素的坐标
        """
        self.x = self.dx + int(rx)
        self.y = self.dy + int(ry)
        return self.x, self.y

    def __str__(self):
        return f"Offset({self.dx}, {self.dy})"


class Position:
    def __init__(self, value=None, align: str | Offset = 'center', region=None):
        """
        定义画面上的位置对象，根据文字或图片定义位置

        @param value:
        @param align:
        @param region: 可以是AnchorRegion对象，由Step类自动设置Anchor的区域
        """
        self.value = value
        self.align = align
        self.region: RectRegion | AnchorRegion | None | tuple = region

    def get_point(self, win_frame: "Frame"):
        """
        获取位置在画面上的相对坐标
        @param win_frame:
        @return:
        """
        if isinstance(self.value, str):
            pos = win_frame.get_text_position(self.value)
            return pos
        else:
            return None


class Exist:
    def __init__(self, value: str | list = "查找的文字或列表", type_=None, op="or", region=None,
                 last_time=0, interval=1):
        """
        判断画面中是否存在某个元素的类，如果条件判断过之后，可以通过对象的pos和obj成员变量获取具体的位置和对象信息.
        在last_time时间段内，任意一次查询找到了value，则Exist对象的satisfied方法返回True，否则返回False。默认1s查找一次

        @param type_: 元素的类型，可取值为text或pic，分别表示文字与图片
        @param value: 元素的值的列表
        @param op: 元素值列表存在于画面中的条件是 与还是或
        @param region: 可以通过该参数指定画面区域，则会在当前区域中查找
        @param last_time: 判断的持续时间，取值单位为s，例如10表示10s内一直满足存在判断条件时，则返回True。取值为0则只判断一次
        """
        self.value = value if isinstance(value, list) else [value]
        self.type_ = type_
        self.op = op
        self.region: AnchorRegion | None = region
        self.last_time = last_time
        self.interval = interval
        self.pos = None  # 如果找到目标，则目标所在的区域，xyxy
        self.obj = None

    def parse_region(self, anchor: RectRegion = None, window_size=None):
        if isinstance(self.region, AnchorRegion):
            self.region = self.region.get_region(anchor_region=anchor, window_size=window_size)
        elif isinstance(self.region, RectRegion):
            ...
        else:  # tuple
            ...

    def _satisfied(self, win_frame: "Frame"):
        """
        画面中存在某个元素的条件是否满足，该方法执行后，可以通过对象的pos和obj成员变量获取具体的位置和对象信息

        @param win_frame:
        @return:
        """
        if isinstance(self.region, AnchorRegion):
            logger.debug(
                f"Exist对象的satisfied()方法调用前需要先调用parse_region()方法以设置目标查询区域的参考区域")
            self.region = None
        if self.type_ is None:
            for v in self.value:
                if v.lower().endswith(".png") or v.lower().endswith(".jpg") or v.lower().endswith(".jpeg"):
                    self.type_ = "pic"
                    return self._satisfied(win_frame)
                else:
                    self.type_ = "text"
                    return self._satisfied(win_frame)
            logger.debug(f"判断条件不存在，默认满足条件！")
            return True
        elif self.type_ == "text":
            text = win_frame.get_text(self.region)
            if self.op == "or":
                for v in self.value:
                    if v in text:
                        logger.debug(f"当前画面中找到【{v}】 in 【{text}】")
                        self.pos = win_frame.get_text_position(v, self.region)
                        if self.pos is not None and self.pos[0] is not None:
                            self.pos = tuple([int(i) for i in self.pos])
                        self.obj = v
                        return True
                return False
            else:
                for v in self.value:
                    if v not in text:
                        return False
                return True
        elif self.type_ == "pic":
            if self.region is None:
                pic = win_frame.capture_window_xywh()
            else:
                pic = win_frame.capture_window_xywh()

            if len(self.value) == 1:
                small_pic_path = os.path.abspath(self.value[0])
                pic_w, pic_h = get_pic_size(small_pic_path)
                res = find_pic(small_pic_path, pic)  # 找到的坐标必然为正值
                self.pos = (res[1][0], res[1][1], res[1][0] + pic_w, res[1][1] + pic_h)
                self.pos = tuple([int(i) for i in self.pos])
                self.obj = self.value[0]
                return res[0]
            else:
                if self.op == "or":
                    for v in self.value:
                        small_pic_path = os.path.abspath(v)
                        res = find_pic(small_pic_path, pic)
                        pic_w, pic_h = get_pic_size(small_pic_path)
                        if res[0]:
                            self.obj = v
                            self.pos = (res[1][0], res[1][1], res[1][0] + pic_w, res[1][1] + pic_h)
                            self.pos = tuple([int(i) for i in self.pos])
                            return True
                    return False
                elif self.op == "and":
                    for v in self.value:
                        small_pic_path = os.path.abspath(v)
                        res = find_pic(small_pic_path, pic)
                        if not res[0]:
                            return False
                        self.obj = v
                        self.pos = res[1]
                    return True

    def satisfied(self, win_frame: "Frame", interval=1):
        if interval != 1 and self.interval == 1:
            self.interval = interval
        if self.last_time == 0:
            return self._satisfied(win_frame)
        else:
            for t in range(self.last_time + 1):
                if self._satisfied(win_frame):
                    return True
                time.sleep(self.interval)
            return False

    def __str__(self):
        return f"exist {self.value}"


class NotExist:
    def __init__(self, value: str | list = "查找的文字或列表", type_="text", op="or",
                 region=None,
                 last_time=0):
        """
        在last_time时间内一直找不到value，则NotExist对象的satisfied()方法返回True。任意一次找到则返回False。默认1s查找一次
        """
        self.exist = Exist(value=value, type_=type_, op=op, region=region, last_time=last_time)
        self.pos = self.exist.pos
        self.obj = self.exist.obj
        self.region = self.exist.region

    def satisfied(self, win_frame: "Frame", interval=1):
        return not self.exist.satisfied(win_frame, interval=interval)

    def __str__(self):
        return f"not exist {self.exist.value}"


class Step:
    def __init__(self, op, target, judge=None, condition=None, wait_method=None):
        """
        定义步骤，
        示例1：
        Step("press", "Ctrl+C", "until", Exist("人物"))  # 按Ctrl+C，直到画面中出现"人物"
        示例2：
        Step(None, None, "until", Exist("竹林偷伐者"))  # 相当于一直等待，直到画面中出现"竹林偷伐者"，该步骤才能通过，进而执行下一步
        示例3：
        Step("long-press-5", "R", None, None)  # 长按R键5s

        :param op: 操作类型，可取值"press", "double-press","left-click", "long-press-{i}",None, "right-click"等，
        其中"long-press-{i}"表示长按某个键i秒，如示例3所示。None则表示不执行操作，可用于等到condition条件满足时使用，
        op也可以是一个函数或方法，这种情况下，Step.do()将直接调用该方法，此时self.target作为传递给该方法的参数。

        :param target: target可以是一个字符串，表示画面上的文字的位置，也可以是一个图片文件，表示画面中图片的位置，也可以是一个Offset对象，
        表示相对于上一个步骤的判断条件对象的偏移量。

        :param judge: 判断操作完成的类型，可取值为until、if或None，如果为None，则表示不判断
        :param wait_method: 条件不满足时，调用的方法，repeat表示重复执行op，如果取值为None，则等待而不重复操作
        """
        self.op = op

        # 按键操作则为键名，鼠标时间则为鼠标点击位置，鼠标点击位置可以通过Position类或Align定义
        self.target: Position | Offset | str | tuple = target
        self.target_region: RectRegion | None = None

        self.judge = judge
        self.condition: "Exist" = condition  # 如果Exist对象的region变量是一个AnchorRegion，则其anchor自动对应为self.target_region
        self.wait_method = wait_method

        self.pos = None  # 该步骤执行后，until后条件中判断的位置区域会保存在该变量中，便于后续使用，可能是(x1,y1,x2,y2)的形式，也可能是(x,y)的形式
        self.obj = None  # 该步骤执行后，until后条件中判断的对象名会保存在该变量中，便于后续使用
        self.frame: "Frame" | None = None
        self.last_step: Step | None = None
        self.before_action = None  # 本步骤操作前的画面
        self.after_action = None  # 本步骤操作后的画面
        self.window_size = None  # 窗口尺寸

    def pre_do(self, last_step):
        if last_step is None:
            return
        else:
            self.last_step = last_step

        if isinstance(self.target, Offset):
            if last_step.pos is not None:
                self.target.set_anchor_pos(last_step.pos[0], last_step.pos[1])
                self.target_region = RectRegion().set_xywh(self.target.x, self.target.y, 1, 1)
            else:
                logger.warning(f"操作的目标对象为Offset类型，但上一个步骤没有生成对应的参考位置，{last_step.pos=}")
                return
        elif isinstance(self.target, tuple):
            self.target_region = RectRegion().set_xywh(self.target[0], self.target[1], 1, 1)
        elif isinstance(self.target, str):
            if "press" in self.op:  # 说明target是个按键，按键不存在坐标位置
                self.target_region = None
            else:
                # 查找字符串或图片在画面中的位置
                self.target_region = None

    def _action(self, win_frame: "Frame"):
        logger.debug(f"操作：{self.op} {self.target}")
        self.window_size = win_frame.window_size
        if self.op is None:
            return
        elif callable(self.op):  # 如果op是一个可以调用的函数或方法，则直接调用
            if self.target is not None:
                self.op(self.target)
            else:
                self.op()
        elif self.op == "press":
            win_frame.press_key(self.target)
        elif self.op == "double-press":
            sleep_interval = random.randint(1, 10) / 100 + 0.1
            win_frame.press_key(self.target)
            time.sleep(sleep_interval)
            win_frame.press_key(self.target)
        elif self.op.startswith("long-press-"):
            # 长按某个键5s
            last_time = int(self.op.replace("long-press-", ""))
            # post_key(win_frame.window, self.target, last_time=last_time)
            win_frame.long_press_key(self.target, last_time=last_time)
        elif self.op == "left-click" or self.op == "click":
            if isinstance(self.target, Position):
                x, y = self.target.get_point(win_frame)
                win_frame.left_click(x, y)
            elif isinstance(self.target, tuple):
                x, y = self.target
                win_frame.left_click(x, y)
            elif isinstance(self.target, Offset):
                x, y = self.target.x, self.target.y
                win_frame.left_click(x, y)
        elif self.op == "right-click":
            if isinstance(self.target, Position):
                x, y = self.target.get_point(win_frame)
                win_frame.right_click(x, y)
            elif isinstance(self.target, tuple):
                x, y = self.target
                win_frame.right_click(x, y)
            elif isinstance(self.target, Offset):
                x, y = self.target.x, self.target.y
                win_frame.right_click(x, y)

    def do(self, win_frame: "Frame"):
        self.frame = win_frame
        logger.debug(f"执行步骤：{self.__str__()}")
        self.before_action = win_frame.capture_window_xywh()
        self._action(win_frame)
        if self.judge is not None:
            time.sleep(0.5)
            if isinstance(self.condition, Exist):
                self.condition.parse_region(anchor=self.target_region, window_size=win_frame.window_size)
            satisfied = self.condition.satisfied(win_frame)
            while not satisfied:  # 如果结束条件不满足，则继续执行本步骤定义的操作
                if self.wait_method is None:
                    time.sleep(0.5)
                else:
                    self._action(win_frame)
                    time.sleep(1)
                satisfied = self.condition.satisfied(win_frame)
            self.after_action = win_frame.capture_window_xywh()
            self.pos = self.condition.pos
            self.obj = self.condition.obj
        else:
            self.pos = None
            self.obj = None
        logger.debug(f"步骤执行完毕：{self.__str__()}")
        return True

    def __str__(self):
        if self.judge is not None:
            return f"{self.op} {self.target} {self.judge} {self.condition.__str__()}"
        else:
            return f"{self.op} {self.target}"

    def show(self, save_to=None):
        """
        调试时显示当前步骤的逻辑。

        上一步的参考区域为淡绿色实线框
        相对于上一步的偏移为绿色箭头
        查找的区域为红色实线框
        找到的目标为绿色实线框
        操作为红色字体

        :param save_to: 保存的gif图片的路径，必须以.gif结尾
        """
        if "press" in self.op:
            cv2.putText(self.before_action, text=f"{self.op} {self.target}", org=(20, 20),
                        fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=1, color=(255, 0, 0), thickness=2)
        elif "click" in self.op:
            if isinstance(self.target, Offset):
                last_pos = self.last_step.pos[0:2]
                p2 = (int(self.target.x), int(self.target.y))
                cv2.rectangle(self.before_action, pt1=last_pos, pt2=self.last_step.pos[2:4],
                              color=(0, 0, 255), thickness=2, lineType=cv2.LINE_AA)
                cv2.arrowedLine(self.before_action, pt1=last_pos, pt2=p2,
                                color=(0, 255, 0), thickness=2, line_type=cv2.LINE_AA, tipLength=0.2)
                cv2.putText(self.before_action, text=f"{self.op} {self.target}", org=p2,
                            fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=1, color=(0, 255, 0), thickness=2)
            elif isinstance(self.target, tuple):  # Offset类型会在执行过程中转换为tuple
                last_pos = self.last_step.pos[0:2]
                p2 = self.target
                cv2.rectangle(self.before_action, pt1=last_pos, pt2=self.last_step.pos[2:4],
                              color=(0, 0, 255), thickness=2, lineType=cv2.LINE_AA)
                cv2.arrowedLine(self.before_action, pt1=last_pos, pt2=p2,
                                color=(0, 255, 0), thickness=2, line_type=cv2.LINE_AA, tipLength=0.2)
                cv2.putText(self.before_action, text=f"{self.op} {self.target}", org=p2,
                            fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=1, color=(0, 255, 0), thickness=2)

        if isinstance(self.condition, Exist) or isinstance(self.condition, NotExist):
            if self.condition.region is None:  # 如果未指定存在条件的查找位置，则默认全屏查找
                cv2.rectangle(self.after_action, pt1=(0, 0), pt2=self.window_size,
                              color=(255, 0, 0), thickness=2, lineType=cv2.LINE_AA, )
            elif isinstance(self.condition.region, tuple):
                p1 = self.condition.region[0:2]
                p2 = self.condition.region[2:4]
                cv2.rectangle(self.after_action, pt1=p1, pt2=p2,
                              color=(255, 0, 0), thickness=2, lineType=cv2.LINE_AA)

            if isinstance(self.condition.pos, tuple):  # 如果存在条件的位置不为空，说明找到了相关目标，绘制该目标示意图
                p1 = (int(self.condition.pos[0]), int(self.condition.pos[1]))
                p2 = (int(self.condition.pos[2]), int(self.condition.pos[3]))
                w, h = p2[0] - p1[0], p2[1] - p1[1]
                cv2.rectangle(self.after_action, pt1=p1, pt2=p2,
                              color=(0, 255, 0), thickness=2, lineType=cv2.LINE_AA, )
                cv2.putText(self.after_action, text=f"Exist: {self.condition.obj},{w=},{h=}", org=p1,
                            fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=0.6, color=(0, 255, 0), thickness=2)
        # before = draw_element_on_pic(self.before_action, False,
        #                              circle={"center": (0, 0), "radius": 20, "fill": "red", "outline": "yellow"})
        # self.target_region  # 目标区域
        # self.condition.region.get_region()  # 条件区域
        # self.last_step
        # show_pic([self.before_action, self.after_action])
        if save_to is not None:
            from PIL import Image
            if self.after_action is None:
                return
            images = [
                Image.fromarray(self.before_action),
                Image.fromarray(self.after_action)
            ]
            images[0].save(save_to, save_all=True, loop=True, append_images=images[1:], duration=2000)
        return


class Steps:
    def __init__(self, steps, debug=False):
        """
        定义游戏内操作的步骤

        @param debug: 是否为调试模式，如果是调试模式，则会显示每一步的操作示意图
        """
        self.steps = steps

    def run(self, win_frame):
        """
        逐步执行Steps对象中记录的步骤
        """
        last_step: Step | None = None
        i = 1
        for step in self.steps:
            step: Step = step
            step.pre_do(last_step)  # 在步骤执行前进行的操作
            success = step.do(win_frame)  # 执行步骤，可能执行失败，如点击某个按钮但画面上找不到就会失败
            if not success and last_step is not None:  # 如果执行失败，就尝试再次执行上一个步骤
                last_step.do(win_frame)
                success = step.do(win_frame)

            if success:  # 如果当前步骤执行结束，就记录当前步骤
                step.show(save_to=f"step{i}.gif")
                last_step = step
            i += 1


class Frame:
    def __init__(self, title, game_path=None, sim_mode=None, sim_info=None):
        """
        游戏图像识别框架
        例如：frame = Frame("天谕",
                           sim_mode="大漠插件-remote",
                           sim_info={"display": "normal", # 大漠绑定窗口的图色模式，模式参数参考dm.BindWindowEx()方法
                                     "mouse": "normal", # 大漠绑定窗口的鼠标模式
                                     "keypad": "normal", # 大漠绑定窗口的键盘模式
                                     "public": "dx.public.active.api", # 大漠绑定窗口的公用模式
                                     "mode": 0,  # 大漠绑定窗口的总模式
                                     "key": "mmdlljafd8c5cbf193e99306e9eb61ceb5bd44",  # 大漠的注册码
                                     "add_key": "ewdSEKFZP",
                                     "guard_type": "memory4", # 开启大漠防护盾的类型，参考dm.DmGuard()方法
                                     "port": 8765,
                                     })  # 大漠的附加码

        :@param title: 游戏窗口标题，为标题的子字符串，但需要唯一确定的窗口
        :@param game_path: 游戏文件路径，如果不存在游戏窗口，则启动该可执行文件
        :@param sim_mode: 模拟键鼠的方法，默认是python自带的方法，可以取值【大漠插件】/【大漠插件-remote】则会使用大漠插件的模拟键鼠方法，
        使用大漠插件时，可以通过sim_info参数传入大漠插件绑定窗口的模式，【大漠插件-remote】使用restful接口请求远程大漠插件操作程序进行模拟鼠标键盘操作，该参数主要是为了解决
        大漠插件没有64位dll的问题，通过远程实现64位python调用32位大漠插件
        """
        if sim_info is None:
            sim_info = {}
        self.set_window_size_steps = None  # 设置窗口尺寸的步骤
        self.title = title
        self.snapshot = None
        self.window = find_window(title, exact=False)
        self.sim_mode = sim_mode
        if isinstance(self.window, dict) and len(self.window) == 1:
            _ = list(self.window.keys())[0]
            self.window_title = self.window[_]
            self.window = _
        else:
            if len(self.window) > 1:
                _ = list(self.window.keys())[0]
                self.window_title = self.window[_]
                self.window = _
                logger.warning(f"找到多个包含{title}的窗口，默认使用第一个:{self.window_title}")
            else:
                logger.error(f"未找到或找到多个包含{title}的窗口，目前不支持这种情况！")
                sys.exit()
        self.window_size = get_size_of_window(self.window, True)  # 不包括标题栏的大小
        self.udf_steps = {}
        if sim_mode is None:
            pass
        elif sim_mode == "大漠插件":
            logger.debug(f"使用大漠插件进行键鼠操作")
            self.dm = DM(key=sim_info.get("key"), add_key=sim_info.get("add_key"))
            if sim_info.get("guard_type") is not None:
                self.dm.DmGuard(1, sim_info.get("guard_type"))
            self.dm.set_bindmode(sim_info.get("display"), sim_info.get("mouse"),
                                 sim_info.get("keypad"), sim_info.get("public"),
                                 sim_info.get("mode"))
            self.dm.bind_window(self.window)
        elif sim_mode == "大漠插件-remote":
            logger.debug(f"使用大漠插件-remote进行键鼠操作")
            self.dm = DMRemote(key=sim_info.get("key"), add_key=sim_info.get("add_key"), port=sim_info.get("port"))
            if sim_info.get("guard_type") is not None:
                self.dm.DmGuard(1, sim_info.get("guard_type"))
            self.dm.bind_window(self.window, sim_info.get("display"), sim_info.get("mouse"),
                                sim_info.get("keypad"), sim_info.get("public"),
                                sim_info.get("mode"))

        self.map: Map | None = None
        self.chat: Chat | None = None
        self.task: Task | None = None
        self.time_panel: TimePanel | None = None  # 时间面板

    def press_key(self, key):
        if self.sim_mode == "大漠插件" or self.sim_mode == "大漠插件-remote":
            self.dm.press_key(key)
        else:
            post_key(self.window, key)

    def long_press_key(self, key, last_time):
        logger.debug(f"暂不支持")
        if self.sim_mode == "大漠插件" or self.sim_mode == "大漠插件-remote":
            self.dm.press_key(key)
        else:
            post_key(self.window, key)

    def left_click(self, x, y, offset=(0, 0)):
        x = x + offset[0]
        y = y + offset[1]
        if self.sim_mode == "大漠插件" or self.sim_mode == "大漠插件-remote":
            self.dm.left_click(x, y)
        else:
            do_click(x, y, self.window)

    def right_click(self, x, y, offset=(0, 0)):
        x = x + offset[0]
        y = y + offset[1]
        if self.sim_mode == "大漠插件" or self.sim_mode == "大漠插件-remote":
            self.dm.right_click(x, y)
        else:
            do_click(x, y, self.window, right=True)

    def drag_to(self, from_point, to_point, button="left"):
        """
        鼠标拖动，从from_point拖动到to_point，button表示是鼠标哪个键，可取值"left","right","mid"
        """
        if self.sim_mode == "大漠插件" or self.sim_mode == "大漠插件-remote":
            self.dm.drag_to(from_point, to_point, button)
        else:
            logger.warning(f"暂不支持鼠标拖动")

    def get_target_position(self, target, region=None):
        if target.lower().endswith("jpg") or target.lower().endswith("jpeg") or target.lower().endswith(
                "png"):
            res = self.get_pic_position_xyxy(target, region)
        else:
            res = self.get_text_position(target, region)
        return res

    def turn_direction_to(self, target, region=None):
        """
        将角色面向转向某个目标
        """
        res = self.get_target_position(target, region)  # (x,y,x,y)
        if res is None or res[0] is None:
            res = self.get_target_position(target, region)  # (x,y,x,y)
            if res is None or res[0] is None:
                logger.debug(f"未找到目标{target=}")
                return
        x1, y1 = int(self.window_size[0] / 2), int((res[1] + res[3]) / 2)
        x2, y2 = int((res[0] + res[2]) / 2), int((res[1] + res[3]) / 2)
        self.drag_to((x1, y1), (x2, y2), button="right")

    def show_snapshot(self):
        show_pic(self.snapshot)

    def define_set_window_size(self, steps: Steps):
        self.set_window_size_steps = steps

    def define_steps(self, steps_name, steps):
        self.udf_steps.update({steps_name: steps})

    def capture_window_xywh(self, x=0, y=0, w=0, h=0):
        """
        按照xywh四个点的坐标截取屏幕区域的图片
        """
        if w == 0:
            w = self.get_window_size()[0]
        if h == 0:
            h = self.get_window_size()[1]
        x2 = x + w
        y2 = y + h
        self.snapshot = capture_pic_undecorated(self.window, x, y, x2, y2)
        return self.snapshot

    def capture_window_xyxy(self, x1, y1, x2, y2):
        """
        按照xyxy四个点的坐标截取屏幕区域的图片，第一个xy是区域左上角坐标，第二个xy是区域右下角坐标
        """
        self.snapshot = capture_pic_undecorated(self.window, x1, y1, x2, y2)
        return self.snapshot

    def set_window_size(self, width=None, height=None):
        """
        将游戏窗口设置为指定大小
        @return:
        """
        # show_pic(self.capture_window())
        self.run_steps(self.set_window_size_steps)

    def run_steps(self, steps: Steps | str):
        if isinstance(steps, Steps):
            steps.run(self)
        else:
            if self.udf_steps.get(steps) is not None:
                self.run_steps(self.udf_steps.get(steps))
        logger.debug("步骤序列执行完毕")

    def run_steps_forever(self, steps: Steps | str, interval=5):
        """
        无限执行指定按键步骤，常用于卡键盘按键
        """

        while True:
            random_sec = random.randint(1, 10) / 10
            self.run_steps(steps)
            time.sleep(random_sec + interval)
            # logger.debug(f"运行步骤{steps=}")

    def get_text(self, region=None):
        self.snapshot = self.capture_region(region)

        return ocr(self.snapshot, paragraph=True)

    def capture_region(self, region):
        region = self.positive_region(region)

        if region is not None:
            if isinstance(region, tuple):
                pic = self.capture_window_xyxy(*region)
            else:  # RectRegion
                pic = self.capture_window_xyxy(region.x1(), region.y1(), region.x2(), region.y2())
        else:
            pic = self.capture_window_xywh()
        return pic

    def positive_region(self, region):
        if region is None:
            return
        w = self.window_size[0]
        h = self.window_size[1]
        if isinstance(region, tuple):
            x1 = region[0] + w if region[0] < 0 else region[0]
            x2 = region[2] + w if region[2] < 0 else region[2]
            y1 = region[1] + h if region[1] < 0 else region[1]
            y2 = region[3] + h if region[3] < 0 else region[3]
            region = (x1, y1, x2, y2)
        elif isinstance(region, RectRegion):  # RectRegion
            region.transfer_to_positive(width=self.window_size[0], height=self.window_size[1])
        else:  # Region
            region = region.transfer_to_positive(width=self.window_size[0], height=self.window_size[1])
        return region

    def get_text_position(self, text, region=None, method="first"):
        """
        获取给定文字在画面上的位置

        :param text: 需要查找的文字内容
        :param region: 查找文字的区域
        :param method: 可取值"first", "most similar"，表示查找到的第一个或者最相近的。
        """
        pic = self.capture_region(region)

        texts = ocr(pic, paragraph=False)
        x1, y1, x2, y2 = None, None, None, None  # 这是相对于region的坐标
        for line in texts:
            position, content = line
            p1, p2, p3, p4 = position
            text_content, text_probability = content
            if text in text_content:
                x1 = p1[0]
                y1 = p1[1]
                x2 = p3[0]
                y2 = p3[1]
                break

        if isinstance(region, RectRegion):
            rx = region.left
            ry = region.top
        elif region is None:
            rx = 0
            ry = 0
        elif isinstance(region, Region):
            rx = region.x1
            ry = region.y1
        else:
            rx = region[0]
            ry = region[1]

        if x1 is not None:
            x1 = x1 + rx
            x2 = x2 + rx
            y1 = y1 + ry
            y2 = y2 + ry

        return x1, y1, x2, y2

    def get_pic_position_xyxy(self, pic, region=None):
        """
        查找图片在画面中的绝对位置
        """
        screen = self.capture_region(region)
        exist, res = find_pic(pic, screen, 0.65)
        if not exist:
            logger.warning(f"查找的图片可信度较低，返回的图片位置可能错误{pic=}")
        size = get_pic_size(pic)
        x1, y1 = res
        x2, y2 = x1 + size[0], y1 + size[1]
        return x1, y1, x2, y2

    def get_window_size(self, refresh=False):
        if refresh:
            self.window_size = get_size_of_window(self.window)
        return self.window_size

    def init_chat(self, anchor: str | list = None, anchor_region: AnchorRegion | None = None, channels=None):
        if isinstance(anchor, str):  # 说明只有一个参考点
            _ = anchor.lower()
            if _.endswith(".jpg") or _.endswith(".png") or _.endswith("jpeg"):
                res = self.get_pic_position_xyxy(anchor)
            else:
                res = self.get_text_position(anchor)
            chat_region = anchor_region.get_region(anchor_region=res, window_size=self.window_size)
            self.chat = Chat(chat_region, channels)
        else:
            ...

    def init_time(self, region=None):
        if region is not None:
            self.time_panel = TimePanel(region=region, frame=self)

    def get_time(self):
        return self.time_panel.get_time()

    def init_task(self, region=None, find_region=None, anchor: str | None = None, anchor_region=None):
        """
        初始化任务面板

        :param region: 任务面板所处的绝对区域位置，为xyxy表示的四个int数值，如果任务面板位置不变的时候，可以使用该参数指定任务面板位置
        :param find_region: 查找任务面板位置的区域，在该区域内查找任务面板的参考点anchor的位置
        :param anchor: 任务面板参考点，根据该点确定任务面板位置
        :param anchor_region: 指定任务面板相对参考点的位置区域。
        """
        if region is not None:
            self.task = Task(region_xyxy=region)
            self.task.set_frame(self)
        else:
            if anchor is None:
                logger.warning(f"任务面板初始化失败，没有任务面板位置的识别特征")
                return
            elif anchor.lower().endswith("jpg") or anchor.lower().endswith("jpeg") or anchor.lower().endswith(
                    "png"):
                res = self.get_pic_position_xyxy(anchor, find_region)
            else:
                res = self.get_text_position(anchor, find_region)
            if res[0] is None:
                return False
            # res必须是正值的region
            task_region = anchor_region.get_region(anchor_region=res, window_size=self.window_size)
            self.task = Task(task_region, frame=self)
            return True

    def get_task_text(self):
        return self.task.get_task_text(self)

    def show_region(self, region):
        if isinstance(region, RectRegion):
            show_pic(self.capture_window_xyxy(*region.get_xyxy()))


class Region:
    def __init__(self, win_frame: "Frame | None" = None, align="center", width_child=0, height_child=0):
        """
        定义窗口中的某个区域，通过该类可以获取(x,y,w,h)这种类型的相对区域坐标
        示例：
        region = Region(frame)
        r.get_region(align="center", width_child=100, height_child=110)

        示例2：
        region = Region(align="center", width_child=100, height_child=110).get_region(win_frame)

        get_region()方法返回的是区域表示的(x,y,w,h)四个参数的tuple
        """
        self.frame = win_frame
        if self.frame is not None:
            self.width_window, self.height_window = win_frame.get_window_size()
            self.align = None
            self.width_child = None
            self.height_child = None
        else:
            self.width_window, self.height_window = None, None
            self.align = align
            self.width_child = width_child
            self.height_child = height_child
        self.x1 = None
        self.x2 = None
        self.y1 = None
        self.y2 = None

    def get_region(self, win_frame=None, align=None, width_child=None, height_child=None):
        """
        根据传入的区域定义，获取区域的数字范围，默认返回整个窗口区域。例如：
        region.get_region("center", 100, 100)  # 表示窗口中间100*100范围的(x,y,w,h)定义
        return (x,y,w,h)
        """
        if win_frame is not None:
            self.frame = win_frame
            self.width_window, self.height_window = win_frame.get_window_size()
        self.align = align or self.align or "center"
        self.width_child = width_child or self.width_child
        self.height_child = height_child or self.height_child

        if self.width_child is None or self.width_child == 0:  # 如果不指定区域的宽和高，则默认宽和高等于窗口的宽和高
            self.width_child = self.width_window
        if self.height_child is None or self.height_child == 0:
            self.height_child = self.height_window
        if self.align == "center":
            left_pad = int((self.width_window - self.width_child) / 2)
            top_pad = int((self.height_window - self.height_child) / 2)
        elif align == "left":
            left_pad = 0
            top_pad = int((self.height_window - self.height_child) / 2)
        elif align == "right":
            left_pad = self.width_window - self.width_child
            top_pad = int((self.height_window - self.height_child) / 2)
        elif align == "top":
            left_pad = int((self.width_window - self.width_child) / 2)
            top_pad = 0
        elif align == "bottom":
            left_pad = int((self.width_window - self.width_child) / 2)
            top_pad = self.height_window - self.height_child
        else:
            left_pad = 0
            top_pad = 0
        self.x1 = left_pad
        self.y1 = top_pad
        self.x2 = left_pad + self.width_child
        self.y2 = top_pad + self.height_child
        return left_pad, top_pad, self.width_child, self.height_child

    def transfer_to_positive(self, width, height):
        self.width_window = width
        self.height_window = height
        x, y, w, h = self.get_region()
        x2, y2 = x + w, y + h
        return x, y, x2, y2


if __name__ == "__main__":
    # logger.debug(ocr("temp.png", paragraph=False))
    # frame = Frame('天谕-无束缚3D')
    # frame = Frame('舒海云')
    frame = Frame('天谕-无束缚3D幻想网游', sim_mode="大漠插件-remote", sim_info={
        "display": "gdi2",
        "mouse": "dx.mouse.position.lock.api|dx.mouse.position.lock.message|dx.mouse.clip.lock.api|dx.mouse.input.lock.api|dx.mouse.state.api|dx.mouse.api|dx.mouse.cursor",
        "keypad": "windows",
        "public": "",
        "mode": 11,
        "key": "mmdlljafd8c5cbf193e99306e9eb61ceb5bd44",
        "add_key": "ewdSEKFZP",
        "guard_type": "memory4",
        "port": 8765,
    })
    # frame = Frame('记事本')
    # frame = Frame('炎火前哨, 宁梦')
    # r = Region(frame)
    # steps_1 = Steps(
    #     steps=[
    #         # 按Esc键，直到窗口中心区域(317, 532)大小范围内出现文字"游戏设置"
    #         Step("press", "escape", "until", Exist(value="游戏设置", region=r.get_region("center", 317, 532))),
    #
    #         # 单击中心区域的【游戏设置】，直到中心区域(874, 609)出现【分辨率】
    #         Step("left-click", Position("游戏设置", region=r.get_region("center", 874, 609)), "until",
    #              Exist(value="分辨率", region=r.get_region("center", 874, 609))),
    #
    #         # 单机
    #         Step("left-click", Position("分辨率", align=Align(20)), "until", Exist())
    #     ])
    # frame.define_set_window_size(steps=steps_1)
    # frame.set_window_size()

    steps_2 = Steps(
        steps=[
            Step("press", "numpad0", "until", Exist(value='pic1.png', type_='pic'), wait_method="repeat"),
            Step("left-click", Offset(397, 16), "until", Exist("门派"), wait_method="repeat"),
            Step("left-click", Offset(0, 0), "until", Exist("苏澜郡"), wait_method="repeat"),
            Step("left-click", Offset(0, 0), "until", Exist("汐族"), wait_method="repeat"),
        ]
    )
    frame.define_steps("打开苏澜郡声望面板", steps_2)
    # frame.run_steps("打开苏澜郡声望面板")

    steps_重复按键 = Steps(
        steps=[
            # Step("press", "R", None),
            Step("press", "R", "until",
                 NotExist("竹林偷伐者", last_time=20, region=Region(align="center", width_child=600)),
                 wait_method="repeat"),
            Step("double-press", "space", "until",
                 Exist("竹林偷伐者", last_time=60, interval=10, region=Region(align="center", width_child=600)),
                 wait_method="repeat"),
            # Step(None, None, "until",
            #      Exist("草药", last_time=60, interval=5),
            #      wait_method="repeat"),  # 等待场景中出现【草药】
            # Step("right-click", "草药", "until",
            #      Exist(),
            #      wait_method=None
            #      )
        ]
    )

    # frame.run_steps(steps_重复按键)
    frame.run_steps_forever(steps_重复按键, interval=8)
    # frame.capture_window()
    # logger.debug(frame.get_text())



import datetime
# import locale
import io
import os
# import json
import sys
import time
from datetime import date
import logging
from dataset.tushareData import *
import common.fileOperate as fo

import numpy as np
import pandas as pd
import pandas_ta as ta
# import talib  # 下载不成功
import tensorflow as tf
import tushare as ts  # 股票数据下载


def append_MACD_to_stocks(directory: str, code: int):
    """
    读取本地的股票数据，计算以下股票指标并写入文件
    股票指标：\n
    ADX-Average Directional Indicator:平均趋向指标\n
    ADXR-:平均方向指数评估，配合ADX、DMI\n
    AR-BRAR_AR:人气指标\n
    ATR-Average True Ranger:均幅指标\n
    BOLL、BBANDs-Bollinger Bands:布林线指标。结合Trix判断更准确\n
    BR-BRAR_BR:意愿指标\n
    CCI-Commodity Channel Index:顺势指标\n
    CR-BRAR_CR: 该指标比较复杂，且各证券公司客户端数据不一致\n
    DMA-Different of Moving Average:平均差、平行线差指标。中短期\n
    DMA_AMA/DIFMA-10日DMA平均值\n
    DMI-Directional Movement Index:本身不对应任何值，它包括PDI、MDI、ADX和ADXR四个有值的指标\n
    MACD-Moving Average Convergence Divergence:移动平均聚散指标、平滑异同平均\n
    MATRIX-Moving Average of TRIX:通过stockStat['trix_9_sma']获得其值\n
    MAVR-Moving Average of VR:通过stockStat['vr_6_sma']获得其值\n
    MDI-DMI_MDI:DMI的子指标，即 -DI\n
    PDI-DMI_PDI:DMI的子指标，即 +DI\n
    ROC-:变动率指标\n
    RSI-Relative Strength Index:相对强弱指数、相对力度指数\n
    SMA-Simple Moving Average：简单移动平均线\n
    TR-True Ranger:均幅指标\n
    TRIX-Triple Exponentially Smoothed Average:三重指数平滑移动平均指标、三重指数平均线\n
    VR-Volumn Ratio:成交量变异率、成交量比率。中期\n
    WR-Williams Rate:威廉指标\n
    :param directory:
    :param code:
    :return:
    """
    file = os.path.join(directory, code + '.csv')
    df = pd.read_csv(file)

    df = cal_ARBR(df)
    # df['ATR'] = stockStat['atr']
    BOLL = ta.bbands(df['close'])  # df['boll']、df['boll_ub']、df['boll_lb']
    df = df.join(BOLL, how='right')
    cal_CCI(df, n=14)  # 已验证

    del BOLL
    cal_CR(df, n=26)  # 已验证

    # 弥补AR、BR的不足:https://wiki.mbalib.com/wiki/CR%E6%8C%87%E6%A0%87
    df['DMA'] = df['close'].rolling(10).mean() - df['close'].rolling(50).mean()  # stockStat['dma']  # 已验证，也称为DMA_DIF
    df['DMA_AMA'] = df['DMA'].rolling(10).mean()  # 已验证，也称为DMA_DIFMA
    cal_DMI(df)
    # ta.adx()

    # df['DX'] = stockStat['dx']
    cal_KDJ(df, 9)
    cal_MACD(df)  # pandas_ta的macd计算在某些情况下会报下标越界错误，这里自己编写

    cal_WR(df, 10, 6)  # 已验证

    df['ROC'] = ta.roc(df['close'])
    df['RSI'] = ta.rsi(df['close'])
    # df['OBV'] = ta.obv(close=df['close'], volume=df['volume'])
    df['OBV'] = talib.OBV(df['close'], df['volume'])

    # df['TR'] = stockStat['tr']
    df['TRIX'] = talib.TRIX(df['close'], timeperiod=14)
    df['MATRIX'] = talib.MA(df['TRIX'], 9, 0)

    cal_VR(df)  # stockStat['vr']

    # df.drop(['price_change', 'p_change', 'v_ma5', 'v_ma10', 'ma5', 'ma10', 'ma20',
    #          'v_ma20', 'high_delta', 'um', 'low_delta', 'dm', 'pdm', 'pdm_14_ema', 'pdm_14',
    #          'close_-1_s', 'tr', 'atr_14', 'mdm', 'mdm_14_ema', 'pdi_14', 'pdi',
    #          'mdm_14', 'mdi_14', 'mdi', 'dx_14', 'dx',
    #          'dx_6_ema', 'adx', 'adx_6_ema', 'adxr',
    #          'atr',  # 'middle', 'middle_14_sma', 'cci_14','cr', 'cr-ma1', 'cr-ma2', 'cr-ma3',
    #          'close_10_sma', 'close_50_sma', 'dma', 'wr_10', 'wr_6', 'change', 'vr'],
    #         axis=1, inplace=True)

    df.to_csv(file, index=False)


def cal_VR(df, n=26, m=6):
    # TH:=SUM(IF(CLOSE>REF(CLOSE,1),VOL,0),N);
    # TL:=SUM(IF(CLOSE<REF(CLOSE,1),VOL,0),N);
    # TQ:=SUM(IF(CLOSE=REF(CLOSE,1),VOL,0),N);
    # VR:100*(TH*2+TQ)/(TL*2+TQ);
    # MAVR:MA(VR,M);
    ref1 = pd.Series(np.where(df['close'] > df['close'].shift(1), df['volume'], 0))
    th = ref1.rolling(n).sum()
    ref2 = pd.Series(np.where(df['close'] < df['close'].shift(1), df['volume'], 0))
    tl = ref2.rolling(n).sum()
    ref3 = pd.Series(np.where(df['close'] == df['close'].shift(1), df['volume'], 0))
    tq = ref3.rolling(n).sum()
    vr = 100 * (th * 2 + tq) / (tl * 2 + tq)
    df['VR'] = np.array(vr)
    df['VR_MAVR'] = df['VR'].rolling(m).mean()
    return df


def cal_WR(df, n=10, n1=6):
    # WR1:100*(HHV(HIGH,N)-CLOSE)/(HHV(HIGH,N)-LLV(LOW,N));
    # WR2:100*(HHV(HIGH,N1)-CLOSE)/(HHV(HIGH,N1)-LLV(LOW,N1));
    temp = pd.DataFrame()
    temp['HHV'] = df['high'].rolling(n).max()
    temp['LLV'] = df['low'].rolling(n).min()
    df['WR_' + str(n)] = 100 * (temp['HHV'] - df['close']) / (temp['HHV'] - temp['LLV'])
    temp['HHV1'] = df['high'].rolling(n1).max()
    temp['LLV1'] = df['low'].rolling(n1).min()
    df['WR_' + str(n1)] = 100 * (temp['HHV1'] - df['close']) / (temp['HHV1'] - temp['LLV1'])
    return df


def cal_DMI(df, n1=14, n2=6):
    """
    本方法已验证，stockstats中DMI计算的四个参数全部错误
    :param df:
    :param n1:
    :param n2:
    :return:
    """
    '''
    MTR:=SUM(MAX(MAX(HIGH-LOW,ABS(HIGH-REF(CLOSE,1))),ABS(REF(CLOSE,1)-LOW)),N);
    HD :=HIGH-REF(HIGH,1);
    LD :=REF(LOW,1)-LOW;
    DMP:=SUM(IF(HD>0&&HD>LD,HD,0),N);
    DMM:=SUM(IF(LD>0&&LD>HD,LD,0),N);
    PDI: DMP*100/MTR;
    MDI: DMM*100/MTR;
    ADX: MA(ABS(MDI-PDI)/(MDI+PDI)*100,M);
    ADXR:(ADX+REF(ADX,M))/2;
    '''

    HD = df['high'] - df['high'].shift(1)
    LD = df['low'].shift(1) - df['low']

    T1 = df['high'] - df['low']
    T2 = abs(df['high'] - df['close'].shift(1))

    TEMP1 = np.where(T1 >= T2, T1, T2)
    TEMP2 = abs(df['low'] - df['close'].shift(1))
    TEMP = pd.Series(np.where(TEMP1 >= TEMP2, TEMP1, TEMP2))

    TR = TEMP.rolling(n1).sum()

    # df.iloc[(HD > 0) & (HD > LD), 'hd1'] = HD
    # df.iloc[(LD > 0) & (LD > HD), 'ld1'] = LD
    HD1 = pd.Series(np.where(((HD > 0) & (HD > LD)), HD, 0))
    LD1 = pd.Series(np.where(((LD > 0) & (LD > HD)), LD, 0))

    DMP = HD1.rolling(n1).sum()
    DMM = LD1.rolling(n1).sum()

    df['DMI_PDI'] = np.array(DMP / TR * 100)
    df['DMI_MDI'] = np.array(DMM / TR * 100)
    TEMP3 = abs(df['DMI_MDI'] - df['DMI_PDI']) / (df['DMI_MDI'] + df['DMI_PDI']) * 100
    df['DMI_ADX'] = np.array(TEMP3.rolling(n2).mean())
    df['DMI_ADXR'] = (df['DMI_ADX'] + df['DMI_ADX'].shift(n2)) / 2

    return df


def cal_ARBR(df: pd.DataFrame, n: int = 26) -> pd.DataFrame:
    stock = pd.DataFrame()
    stock['HO'] = df['high'] - df['open']
    stock['OL'] = df['open'] - df['low']
    stock['HCY'] = df['high'] - df['close'].shift(1)
    # 百度百科等介绍的指标计算没有取非负值的步骤,百度百科关于很多指标的计算公式都有问题
    stock['HCY'] = np.where(stock['HCY'] < 0, 0, stock['HCY'])  # 将stock['HCY']列小于0的数替换为0，即取非负值
    stock['CYL'] = df['close'].shift(1) - df['low']
    stock['CYL'] = np.where(stock['CYL'] < 0, 0, stock['CYL'])  # 将stock['CYL']列小于0的数替换为0
    df['ARBR_AR'] = stock['HO'].rolling(26).sum() / stock['OL'].rolling(26).sum() * 100  # 已验证
    df['ARBR_BR'] = stock['HCY'].rolling(26).sum() / stock['CYL'].rolling(26).sum() * 100  # 已验证
    return df


def cal_CCI(df, n=14):
    """
    计算CCI指标，已验证。
    pandas_ta的cci指标计算结果错误，stockstats的cci指标有警告，另外，
    百度百科里的cci指标计算公式错误，主要是close和TYP弄混了，其次
    # MD = (MA - TYP).abs().rolling(n).mean()计算的MD是不对的，因为MA不应参与rolling,
    而应该与当前MD的索引一致
    :param df:
    :param n:
    :return:
    """
    # TYP: = (HIGH + LOW + CLOSE) / 3;
    # CCI: (TYP - MA(TYP, N)) / (0.015 * AVEDEV(TYP, N));
    TYP = (df['high'] + df['low'] + df['close']) / 3
    MA = TYP.rolling(n).mean()
    cci = []
    for i in range(len(MA)):
        ma = MA.iloc[i]
        if i < n - 1:
            cci.append(np.NaN)
            continue
        else:
            sum = 0
            for j in range(n):
                sum = abs(TYP.iloc[i - j] - ma) + sum
            md = sum / n
        cci.append((TYP.iloc[i] - ma) / (0.015 * md))

    df['CCI'] = cci

    return df


def cal_CR(df, n=26):
    """
    结果已验证，stockstats的cr计算结果错误
    MID:=REF(HIGH+LOW,1)/2;
    CR:SUM(MAX(0,HIGH-MID),N)/SUM(MAX(0,MID-LOW),N)*100;
    MA1:REF(MA(CR,M1),M1/2.5+1);
    MA2:REF(MA(CR,M2),M2/2.5+1);
    MA3:REF(MA(CR,M3),M3/2.5+1);
    MA4:REF(MA(CR,M4),M4/2.5+1);
    :param df:
    :param n:
    :return:
    """
    MID = ((df['high'] + df['low']) / 2).shift(1)
    P1 = df['high'] - MID
    P1 = np.where(P1 < 0, 0, P1)  # 将stock['HCY']列小于0的数替换为0，即取非负值
    P1 = pd.Series(P1)
    P2 = MID - df['low']
    P2 = np.where(P2 < 0, 0, P2)
    P2 = pd.Series(P2)
    CR = P1.rolling(n).sum() / P2.rolling(n).sum() * 100  # 直接使用df['CR'] = CR，结果df['CR']没有数据，全是空值
    df['CR'] = np.array(CR)
    return df


def cal_MACD(df, fast_ma=12, slow_ma=26, ma=9):
    """
    已验证
    中长期投资者首选参考指标，该指标具有滞后性，当行情迅速大幅涨跌时，不适用。\n

    :param df: original dataframe, index = 'date',
                columns ='open','high','close','low','volume','amount'
    :param fast_ma: fast period
    :param slow_ma: slow period
    :param ma: moving average
    :return: DateFrame
    """
    df['MACD_DIFF'] = df['close'].ewm(adjust=False, alpha=2 / (fast_ma + 1), ignore_na=True).mean() - \
                      df['close'].ewm(adjust=False, alpha=2 / (slow_ma + 1), ignore_na=True).mean()
    df['MACD_DEA'] = df['MACD_DIFF'].ewm(adjust=False, alpha=2 / (ma + 1), ignore_na=True).mean()
    df['MACD'] = 2 * (df['MACD_DIFF'] - df['MACD_DEA'])
    return df


def cal_KDJ(df, n=9, ksgn='close'):
    """
    【输入】
        df, pd.dataframe格式数据源
        n，时间长度
        ksgn，列名，一般是：close收盘价
    【输出】    
        df, pd.dataframe格式数据源,
        增加了一栏：_{n}，输出数据
    """
    lowList = df['low'].rolling(n).min()
    lowList.fillna(value=df['low'].expanding().min(), inplace=True)
    highList = df['high'].rolling(n).max()
    highList.fillna(value=df['high'].expanding().max(), inplace=True)
    rsv = (df[ksgn] - lowList) / (highList - lowList) * 100

    df['KDJ_K'] = rsv.ewm(com=2).mean()  # 指数平滑移动平均线
    df['KDJ_D'] = df['KDJ_K'].ewm(com=2).mean()  # Exponentially Weighted Moving-Average

    df['KDJ_J'] = 3.0 * df['KDJ_K'] - 2.0 * df['KDJ_D']
    return df



import requests
from yangke.base import web_available, get_temp_para, pic2base64
import os, socket, urllib3


def get_token(tries=0, client_id="fmt3q4GOHR5lyzjGb78I9Otr", client_secret="DHcIclrolnmPnMFOyf6mjnZ6V2Qda76o"):
    """
    获取access_token，因为百度Unit智能对话定制与服务平台访问需要access_token才能使用，具体使用方法参见：
    http://ai.baidu.com/ai-doc/REFERENCE/Ck3dwjhhu

    已申请的应用:

    应用名称        API Key                          Secret Key
    yk_ocr        fmt3q4GOHR5lyzjGb78I9Otr         DHcIclrolnmPnMFOyf6mjnZ6V2Qda76o     通用API，几乎包含百度所有功能
    智能问答       KBK6CpyAIaxcfURZiDTPHFyv         bwvENVUj7XtXqGVGyYieqH8F20KhKRjw

    :param tries: 尝试次数
    :param client_id: 应用的API Key
    :param client_secret: 应用的Secret Key
    :return: 返回百度的access_token
    """

    # client_id为官网获取的API Key，client_secret 为官网获取的Secret Key
    def _get_token():
        host = f'https://aip.baidubce.com/oauth/2.0/token?grant_type=client_credentials&' \
               f'client_id={client_id}&' \
               f'client_secret={client_secret}'
        try:
            resp = requests.get(host)
            if resp:
                print(resp.json())
                acc_token = resp.json().get('access_token')  # 一般access_token一个月有效期
                return acc_token
            else:
                if tries < 3:  # 如果失败，则重试3次
                    return get_token(tries + 1)
                if web_available():
                    raise Exception("向百度请求access_token参数失败，请核对百度相关接口是否已经更改！")
                else:
                    raise Exception("网络似乎没有连接，请检查重试！")
        except (requests.exceptions.ConnectionError, socket.gaierror, urllib3.exceptions.NewConnectionError,
                urllib3.exceptions.MaxRetryError) as e:
            raise Exception("网络似乎没有连接，请检查重试！")

    access_token = get_temp_para('access_token', expires_in=2500000, get_func_or_default_value=_get_token)
    return access_token


def ocr(image: str = None, method: str = "general"):
    """
    百度文字识别，参考https://cloud.baidu.com/doc/OCR/s/vk3h7y58v

    :param image: 包含文字的图片
    :param method: 识别类型，accurate_basic/accurate/general_basic/general/... 具体参见上述百度帮助页面
    :return:
    """

    image_base64 = pic2base64(image)
    access_token = get_token(client_id="fmt3q4GOHR5lyzjGb78I9Otr", client_secret="DHcIclrolnmPnMFOyf6mjnZ6V2Qda76o")
    url = f'https://aip.baidubce.com/rest/2.0/ocr/v1/{method}?access_token=' + access_token  # 机器人对话API，沙盒环境
    # url = 'https://aip.baidubce.com/rpc/2.0/unit/bot/chat?access_token=' + access_token  # 技能对话API

    post_data = {
        "image": image_base64,
        "language_type": "CHN_ENG",  # auto_detect/CHN_ENG/ENG/
        "detect_direction": "true",
        "paragraph": "true",
        "probability": "true",
    }
    post_data = str(post_data).replace("'", '"').encode('utf-8')

    headers = {'content-type': 'application/x-www-form-urlencoded'}
    response = requests.post(url, data=post_data, headers=headers)
    result = []  # 答案的列表
    if response:
        print(response.json())
        response_list = response.json().get('result').get('response_list')
        for res in response_list:
            action = res.get('action_list')[0]  # action_list是一个列表，这里去列表第一项，如果以后遇到多答案的问题，再更新
            result.append(action.get('say'))

    return result


if __name__ == "__main__":
    ocr(r"C:\Users\YangKe\Pictures\2.png")



# encoding: utf-8
g_lunar_month_day = [
    0x00752, 0x00ea5, 0x0ab2a, 0x0064b, 0x00a9b, 0x09aa6, 0x0056a, 0x00b59, 0x04baa, 0x00752,  # 1901 ~ 1910
    0x0cda5, 0x00b25, 0x00a4b, 0x0ba4b, 0x002ad, 0x0056b, 0x045b5, 0x00da9, 0x0fe92, 0x00e92,  # 1911 ~ 1920
    0x00d25, 0x0ad2d, 0x00a56, 0x002b6, 0x09ad5, 0x006d4, 0x00ea9, 0x04f4a, 0x00e92, 0x0c6a6,  # 1921 ~ 1930
    0x0052b, 0x00a57, 0x0b956, 0x00b5a, 0x006d4, 0x07761, 0x00749, 0x0fb13, 0x00a93, 0x0052b,  # 1931 ~ 1940
    0x0d51b, 0x00aad, 0x0056a, 0x09da5, 0x00ba4, 0x00b49, 0x04d4b, 0x00a95, 0x0eaad, 0x00536,  # 1941 ~ 1950
    0x00aad, 0x0baca, 0x005b2, 0x00da5, 0x07ea2, 0x00d4a, 0x10595, 0x00a97, 0x00556, 0x0c575,  # 1951 ~ 1960
    0x00ad5, 0x006d2, 0x08755, 0x00ea5, 0x0064a, 0x0664f, 0x00a9b, 0x0eada, 0x0056a, 0x00b69,  # 1961 ~ 1970
    0x0abb2, 0x00b52, 0x00b25, 0x08b2b, 0x00a4b, 0x10aab, 0x002ad, 0x0056d, 0x0d5a9, 0x00da9,  # 1971 ~ 1980
    0x00d92, 0x08e95, 0x00d25, 0x14e4d, 0x00a56, 0x002b6, 0x0c2f5, 0x006d5, 0x00ea9, 0x0af52,  # 1981 ~ 1990
    0x00e92, 0x00d26, 0x0652e, 0x00a57, 0x10ad6, 0x0035a, 0x006d5, 0x0ab69, 0x00749, 0x00693,  # 1991 ~ 2000
    0x08a9b, 0x0052b, 0x00a5b, 0x04aae, 0x0056a, 0x0edd5, 0x00ba4, 0x00b49, 0x0ad53, 0x00a95,  # 2001 ~ 2010
    0x0052d, 0x0855d, 0x00ab5, 0x12baa, 0x005d2, 0x00da5, 0x0de8a, 0x00d4a, 0x00c95, 0x08a9e,  # 2011 ~ 2020
    0x00556, 0x00ab5, 0x04ada, 0x006d2, 0x0c765, 0x00725, 0x0064b, 0x0a657, 0x00cab, 0x0055a,  # 2021 ~ 2030
    0x0656e, 0x00b69, 0x16f52, 0x00b52, 0x00b25, 0x0dd0b, 0x00a4b, 0x004ab, 0x0a2bb, 0x005ad,  # 2031 ~ 2040
    0x00b6a, 0x04daa, 0x00d92, 0x0eea5, 0x00d25, 0x00a55, 0x0ba4d, 0x004b6, 0x005b5, 0x076d2,  # 2041 ~ 2050
    0x00ec9, 0x10f92, 0x00e92, 0x00d26, 0x0d516, 0x00a57, 0x00556, 0x09365, 0x00755, 0x00749,  # 2051 ~ 2060
    0x0674b, 0x00693, 0x0eaab, 0x0052b, 0x00a5b, 0x0aaba, 0x0056a, 0x00b65, 0x08baa, 0x00b4a,  # 2061 ~ 2070
    0x10d95, 0x00a95, 0x0052d, 0x0c56d, 0x00ab5, 0x005aa, 0x085d5, 0x00da5, 0x00d4a, 0x06e4d,  # 2071 ~ 2080
    0x00c96, 0x0ecce, 0x00556, 0x00ab5, 0x0bad2, 0x006d2, 0x00ea5, 0x0872a, 0x0068b, 0x10697,  # 2081 ~ 2090
    0x004ab, 0x0055b, 0x0d556, 0x00b6a, 0x00752, 0x08b95, 0x00b45, 0x00a8b, 0x04a4f, ]

# 农历数据 每个元素的存储格式如下：
#    12~7         6~5    4~0
#  离元旦多少天  春节月  春节日
#####################################################################################
g_lunar_year_day = [
    0x18d3, 0x1348, 0x0e3d, 0x1750, 0x1144, 0x0c39, 0x15cd, 0x1042, 0x0ab6, 0x144a,  # 1901 ~ 1910
    0x0ebe, 0x1852, 0x1246, 0x0cba, 0x164e, 0x10c3, 0x0b37, 0x14cb, 0x0fc1, 0x1954,  # 1911 ~ 1920
    0x1348, 0x0dbc, 0x1750, 0x11c5, 0x0bb8, 0x15cd, 0x1042, 0x0b37, 0x144a, 0x0ebe,  # 1921 ~ 1930
    0x17d1, 0x1246, 0x0cba, 0x164e, 0x1144, 0x0bb8, 0x14cb, 0x0f3f, 0x18d3, 0x1348,  # 1931 ~ 1940
    0x0d3b, 0x16cf, 0x11c5, 0x0c39, 0x15cd, 0x1042, 0x0ab6, 0x144a, 0x0e3d, 0x17d1,  # 1941 ~ 1950
    0x1246, 0x0d3b, 0x164e, 0x10c3, 0x0bb8, 0x154c, 0x0f3f, 0x1852, 0x1348, 0x0dbc,  # 1951 ~ 1960
    0x16cf, 0x11c5, 0x0c39, 0x15cd, 0x1042, 0x0a35, 0x13c9, 0x0ebe, 0x17d1, 0x1246,  # 1961 ~ 1970
    0x0d3b, 0x16cf, 0x10c3, 0x0b37, 0x14cb, 0x0f3f, 0x1852, 0x12c7, 0x0dbc, 0x1750,  # 1971 ~ 1980
    0x11c5, 0x0c39, 0x15cd, 0x1042, 0x1954, 0x13c9, 0x0e3d, 0x17d1, 0x1246, 0x0d3b,  # 1981 ~ 1990
    0x16cf, 0x1144, 0x0b37, 0x144a, 0x0f3f, 0x18d3, 0x12c7, 0x0dbc, 0x1750, 0x11c5,  # 1991 ~ 2000
    0x0bb8, 0x154c, 0x0fc1, 0x0ab6, 0x13c9, 0x0e3d, 0x1852, 0x12c7, 0x0cba, 0x164e,  # 2001 ~ 2010
    0x10c3, 0x0b37, 0x144a, 0x0f3f, 0x18d3, 0x1348, 0x0dbc, 0x1750, 0x11c5, 0x0c39,  # 2011 ~ 2020
    0x154c, 0x0fc1, 0x0ab6, 0x144a, 0x0e3d, 0x17d1, 0x1246, 0x0cba, 0x15cd, 0x10c3,  # 2021 ~ 2030
    0x0b37, 0x14cb, 0x0f3f, 0x18d3, 0x1348, 0x0dbc, 0x16cf, 0x1144, 0x0bb8, 0x154c,  # 2031 ~ 2040
    0x0fc1, 0x0ab6, 0x144a, 0x0ebe, 0x17d1, 0x1246, 0x0cba, 0x164e, 0x1042, 0x0b37,  # 2041 ~ 2050
    0x14cb, 0x0fc1, 0x18d3, 0x1348, 0x0dbc, 0x16cf, 0x1144, 0x0a38, 0x154c, 0x1042,  # 2051 ~ 2060
    0x0a35, 0x13c9, 0x0e3d, 0x17d1, 0x11c5, 0x0cba, 0x164e, 0x10c3, 0x0b37, 0x14cb,  # 2061 ~ 2070
    0x0f3f, 0x18d3, 0x12c7, 0x0d3b, 0x16cf, 0x11c5, 0x0bb8, 0x154c, 0x1042, 0x0ab6,  # 2071 ~ 2080
    0x13c9, 0x0e3d, 0x17d1, 0x1246, 0x0cba, 0x164e, 0x10c3, 0x0bb8, 0x144a, 0x0ebe,  # 2081 ~ 2090
    0x1852, 0x12c7, 0x0d3b, 0x16cf, 0x11c5, 0x0c39, 0x154c, 0x0fc1, 0x0a35, 0x13c9,  # 2091 ~ 2100
]

# ==================================================================================

from datetime import date, datetime, timedelta
from yangke.web.webQA import *
import re
import sxtwl

# 开始年份
START_YEAR = 1901

month_DAY_BIT = 12
month_NUM_BIT = 13

# 　todo：正月初一 == 春节   腊月二十九/三十 == 除夕
yuefeng = ["正月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "冬月", "腊月"]
riqi = ["初一", "初二", "初三", "初四", "初五", "初六", "初七", "初八", "初九", "初十",
        "十一", "十二", "十三", "十四", "十五", "十六", "十七", "十八", "十九", "廿十",
        "廿一", "廿二", "廿三", "廿四", "廿五", "廿六", "廿七", "廿八", "廿九", "三十"]

xingqi = ["星期一", "星期二", "星期三", "星期四", "星期五", "星期六", "星期日"]

tiangan = ["甲", "乙", "丙", "丁", "戊", "己", "庚", "辛", "壬", "癸"]
dizhi = ["子", "丑", "寅", "卯", "辰", "巳", "午", "未", "申", "酉", "戌", "亥"]
shengxiao = ["鼠", "牛", "虎", "兔", "龙", "蛇", "马", "羊", "猴", "鸡", "狗", "猪"]

sFtv = [  # 阳历节日 * 表示国家法定假日
    "0101#元旦节#*",
    "0202#世界湿地日#",
    "0210#国际气象节#",
    "0214#情人节#",
    "0301#国际海豹日#",
    "0303#全国爱耳日#",
    "0305#学雷锋纪念日#",
    "0308#妇女节#",
    "0312#植树节# #孙中山逝世纪念日#",
    "0314#国际警察日#",
    "0315#消费者权益日#",
    "0317#中国国医节# #国际航海日#",
    "0321#世界森林日# #消除种族歧视国际日# #世界儿歌日#",
    "0322#世界水日#",
    "0323#世界气象日#",
    "0324#世界防治结核病日#",
    "0325#全国中小学生安全教育日#",
    "0330#巴勒斯坦国土日#",
    "0401#愚人节# #全国爱国卫生运动月(四月)# #税收宣传月(四月)#",
    "0404/0405/0406#清明节#<谷雨>+15*",  # 虽为农历的节气，但每年必然是这三天之一，谷雨后的第15天
    "0407#世界卫生日#",
    "0422#世界地球日#",
    "0423#世界图书和版权日#",
    "0424#亚非新闻工作者日#",
    "0501#劳动节#*",
    "0504#青年节#",
    "0505#碘缺乏病防治日#",
    "0508#世界红十字日#",
    "0512#国际护士节#",
    "0515#国际家庭日#",
    "0517#国际电信日#",
    "0518#国际博物馆日#",
    "0520#全国学生营养日#",
    "0523#国际牛奶日#",
    "0531#世界无烟日#",
    "0601#国际儿童节#",
    "0605#世界环境保护日#",
    "0606#全国爱眼日#",
    "0617#防治荒漠化和干旱日#",
    "0623#国际奥林匹克日#",
    "0625#全国土地日#",
    "0626#国际禁毒日#",
    "0701#中国共·产党诞辰# #香港回归纪念日# #世界建筑日#",
    "0702#国际体育记者日#",
    "0707#抗日战争纪念日#",
    "0711#世界人口日#",
    "0730#非洲妇女日#",
    "0801#建军节#",
    "0808#中国男子节(爸爸节)#",
    "0815#抗日战争胜利纪念#",
    "0908#国际扫盲日# #国际新闻工作者日#",
    "0909#毛·泽东逝世纪念#",
    "0910#中国教师节#",
    "0914#世界清洁地球日#",
    "0916#国际臭氧层保护日#",
    "0918#九·一八事变纪念日#",
    "0920#国际爱牙日#",
    "0927#世界旅游日#",
    "0928#孔子诞辰#",
    "1001#国庆节# #世界音乐日# #国际老人节#*",
    "1002#国庆节假日# #国际和平与民主自由斗争日#",
    "1003#国庆节假日#",
    "1004#世界动物日#",
    "1006#老人节#",
    "1008#全国高血压日# #世界视觉日#",
    "1009#世界邮政日# #万国邮联日#",
    "1010#辛亥革命纪念日# #世界精神卫生日#",
    "1013#世界保健日# #国际教师节#",
    "1014#世界标准日#",
    "1015#国际盲人节(白手杖节)#",
    "1016#世界粮食日#",
    "1017#世界消除贫困日#",
    "1022#世界传统医药日#",
    "1024#联合国日#",
    "1031#世界勤俭日#",
    "1107#十月社会主义革命纪念日#",
    "1108#中国记者日#",
    "1109#全国消防安全宣传教育日#",
    "1110#世界青年节#",
    "1111#国际科学与和平周(本日所属的一周)#",
    "1112#孙中山诞辰纪念日#",
    "1114#世界糖尿病日#",
    "1116#国际宽容日#",
    "1117#国际大学生节# #世界学生节#",
    "1120#彝族年#",
    "1121#彝族年# #世界问候日# #世界电视日#",
    "1122#彝族年#",
    "1129#国际声援巴勒斯坦人民国际日#",
    "1201#世界艾滋病日#",
    "1203#世界残疾人日#",
    "1205#国际经济和社会发展志愿人员日#",
    "1208#国际儿童电视日#",
    "1209#世界足球日#",
    "1210#世界人权日#",
    "1212#西安事变纪念日#",
    "1213#南京大屠杀(1937年)纪念日#",
    "1220#澳门回归纪念#",
    "1221#国际篮球日#",
    "1224#平安夜#",
    "1225#圣诞节#",
    "1226#毛·泽东诞辰纪念日#"
]

lFtv = [  # 农历节日
    "0101#春节#*",
    "0115#元宵节#",
    "0202#春龙节",
    # "0314#清明节#", #每年不一样，此为2012年，事实上为公历节日
    "0505#端午节#*",
    "0707#七夕情人节#",
    "0715#中元节#",
    "0815#中秋节#*",
    "0909#重阳节#",
    "1208#腊八节#",
    "1223#小年#",
    "1229/1230#除夕#"  # 每年不一样，可能是1229或1230，但一定是岁末最后一天，即春节前一天
]


def change_year(num):
    dx = ["零", "一", "二", "三", "四", "五", "六", "七", "八", "九", "十"]
    tmp_str = ""
    # 将年份 转换为字符串，然后进行遍历字符串 ，将字符串中的数字转换为中文数字
    for i in str(num):
        tmp_str += dx[int(i)]
    return tmp_str


# 获取星期
def week_str(tm):
    return xingqi[tm.weekday()]


# 获取天数
def lunar_day(day):
    return riqi[(day - 1) % 30]


def lunar_day1(month, day):
    if day == 1:
        return lunar_month(month)
    else:
        return riqi[day - 1]


# 判断是否是闰月
def lunar_month(month):
    leap = (month >> 4) & 0xf
    m = month & 0xf
    month = yuefeng[(m - 1) % 12]
    if leap == m:
        month = "闰" + month
    return month


# 求什么年份，中国农历的年份和 什么生肖年
def lunar_year(year):
    return tiangan[(year - 4) % 10] + dizhi[(year - 4) % 12] + '[' + shengxiao[(year - 4) % 12] + ']'


# 返回：
# a b c
# 闰几月，该闰月多少天 传入月份多少天
def lunar_month_days(lunar_year, lunar_month):
    if (lunar_year < START_YEAR):
        return 30

    leap_month, leap_day, month_day = 0, 0, 0  # 闰几月，该月多少天 传入月份多少天

    tmp = g_lunar_month_day[lunar_year - START_YEAR]

    if tmp & (1 << (lunar_month - 1)):
        month_day = 30
    else:
        month_day = 29

        # 闰月
    leap_month = (tmp >> month_NUM_BIT) & 0xf
    if leap_month:
        if (tmp & (1 << month_DAY_BIT)):
            leap_day = 30
        else:
            leap_day = 29

    return (leap_month, leap_day, month_day)


# 算农历日期
# 返回的月份中，高4bit为闰月月份，低4bit为其它正常月份
def get_ludar_date(tm):
    year, month, day = tm.year, 1, 1
    code_data = g_lunar_year_day[year - START_YEAR]
    days_tmp = (code_data >> 7) & 0x3f
    chunjie_d = (code_data >> 0) & 0x1f
    chunjie_m = (code_data >> 5) & 0x3
    span_days = (tm - datetime(year, chunjie_m, chunjie_d)).days
    # print("span_day: ", days_tmp, span_days, chunjie_m, chunjie_d)

    # 日期在该年农历之后
    if (span_days >= 0):
        (leap_month, foo, tmp) = lunar_month_days(year, month)
        while span_days >= tmp:
            span_days -= tmp
            if (month == leap_month):
                (leap_month, tmp, foo) = lunar_month_days(year, month)  # 注：tmp变为闰月日数
                if (span_days < tmp):  # 指定日期在闰月中
                    month = (leap_month << 4) | month
                    break
                span_days -= tmp
            month += 1  # 此处累加得到当前是第几个月
            (leap_month, foo, tmp) = lunar_month_days(year, month)
        day += span_days
        return year, month, day
        # 倒算日历
    else:
        month = 12
        year -= 1
        (leap_month, foo, tmp) = lunar_month_days(year, month)
        while abs(span_days) >= tmp:
            span_days += tmp
            month -= 1
            if (month == leap_month):
                (leap_month, tmp, foo) = lunar_month_days(year, month)
                if (abs(span_days) < tmp):  # 指定日期在闰月中
                    month = (leap_month << 4) | month
                    break
                span_days += tmp
            (leap_month, foo, tmp) = lunar_month_days(year, month)
        day += (tmp + span_days)  # 从月份总数中倒扣 得到天数
        return year, month, day


# 打印 某个时间的农历
def _show_month(tm):
    (year, month, day) = get_ludar_date(tm)
    print("%d年%d月%d日" % (tm.year, tm.month, tm.day), week_str(tm), end='')
    print("\t农历 %s年 %s年%s%s " % (lunar_year(year), change_year(year), lunar_month(month), lunar_day(day)))  # 根据数组索引确定
    return lunar_year(year), change_year(year), lunar_month(month), lunar_day(day)


# 判断输入的数据是否符合规则
def show_month(day_datetime: datetime.date = date.today()):
    """
    返回指定的年月日的农历日期

    :param day_datetime
    :return:
    """
    if day_datetime.year > 2100 or day_datetime.year < 1901:
        return
    if day_datetime.month > 13 or day_datetime.month < 1:
        return

    tmp = datetime(day_datetime.year, day_datetime.month, day_datetime.day)
    lu_year, ch_year, lu_month, lu_day = _show_month(tmp)
    return lu_year, ch_year, lu_month, lu_day


def _is_holiday(day_datetime: datetime.date = date.today(), ftv_type='solar') -> (bool, int):
    """
    判断是否是法定 节假日，只有法定放假的节日才返回True。
    法定假日有：
    元旦节 -> 阳历1月1日
    春节 -> 阴历1月1日，由除夕开始放假，即阴历1月1日的前一天开始放假
    清明节 -> 阳历4月4日/4月5日/4月6日，每年不确定
    劳动节 -> 阳历5月1日
    端午节 -> 阴历5月5日
    中秋节 -> 阴历8月15日
    国庆节 -> 阳历10月1日

    :param day_datetime: 需要查询的日期
    :param ftv_type: 节日类型，分为阳历节日和阴历日期，分别取值'solar', 'lunar'
    :return: 是否节假日，置信度
    """
    result = False
    confidence = 0
    md = f"{day_datetime.month:>02}{day_datetime.day:>02}"  # 右对齐，长度为2，不足则在左边补0
    ftv_spe = sFtv if ftv_type == 'solar' else lFtv
    for item in ftv_spe:
        if '*' not in item:  # 如果该节日不是假期，则直接忽略该节日
            continue
        ftv = item.split("#")[0]  # 获取节日的日期
        if '/' in ftv:  # 说明节日日期不确定，例如清明节，母亲节等，因为这些不定节日中只有清明节是法定假日，这里只需判断清明节
            ds = ftv.split('/')
            if md in ds:
                # 说明可能是假日，继续判断是否是节日，清明节的日期不定
                if item.split("#")[1] == "清明节":
                    ques = f'{day_datetime.year}年{item.split("#")[1]}是几月几号？'  # 这里直接询问百度机器人
                    ans = ask_baidu(ques)[0]
                    ans = re.findall(".*(.)月(.)日.*", ans)[0]  # 正则表达式匹配"月"和"日"之前的字符
                    print(ans)
                    if md == f"{ans[0]:>02}{ans[1]:>02}":
                        result = True
                        confidence = 0.8
                        break
            else:
                result = False
                confidence = 1
        else:
            if md == ftv:  # 日期==节日日期
                result = True
                confidence = 1
                break

    return result, confidence


def is_holiday(day_datetime: datetime.date = date.today()) -> (bool, str):
    """
    判断日期是否是国家法定假日

    :param day_datetime: 日期
    :return: (bool, str) 是否假日，判断依据-可靠度
    """
    if _is_holiday(day_datetime, 'solar')[0] or _is_holiday(day_datetime, 'lunar')[0]:
        return True
    else:
        return False


def get_qingming(year: int):
    """
    获得指定年份清明节的日期，year的取值范围为1990年到2099年，超范围本算法不准确
    """
    qm = {'1990': '5', '1991': '5', '1992': '4', '1993': '5', '1994': '5', '1995': '5', '1996': '4',
          '1997': '5', '1998': '5', '1999': '5', '2000': '4', '2001': '5', '2002': '5', '2003': '5',
          '2004': '4', '2005': '5', '2006': '5', '2007': '5', '2008': '4', '2009': '4', '2010': '5',
          '2011': '5', '2012': '4', '2013': '4', '2014': '5', '2015': '5', '2016': '4', '2017': '4',
          '2018': '5', '2019': '5', '2020': '4', '2021': '4', '2022': '5', '2023': '5', '2024': '4',
          '2025': '4', '2026': '5', '2027': '5', '2028': '4', '2029': '4', '2030': '5', '2031': '5',
          '2032': '4', '2033': '4', '2034': '5', '2035': '5', '2036': '4', '2037': '4', '2038': '5',
          '2039': '5', '2040': '4', '2041': '4', '2042': '4', '2043': '5', '2044': '4', '2045': '4',
          '2046': '4', '2047': '5', '2048': '4', '2049': '4', '2050': '4', '2051': '5', '2052': '4',
          '2053': '4', '2054': '4', '2055': '5', '2056': '4', '2057': '4', '2058': '4', '2059': '5',
          '2060': '4', '2061': '4', '2062': '4', '2063': '5', '2064': '4', '2065': '4', '2066': '4',
          '2067': '5', '2068': '4', '2069': '4', '2070': '4', '2071': '5', '2072': '4', '2073': '4',
          '2074': '4', '2075': '4', '2076': '4', '2077': '4', '2078': '4', '2079': '4', '2080': '4',
          '2081': '4', '2082': '4', '2083': '4', '2084': '4', '2085': '4', '2086': '4', '2087': '4',
          '2088': '4', '2089': '4', '2090': '4', '2091': '4', '2092': '4', '2093': '4', '2094': '4',
          '2095': '4', '2096': '4', '2097': '4', '2098': '4', '2099': '4'}
    # qm字典数据的获得方法：
    """
    import re
    qingming_dict = {}
    for year in range(1990, 2100):
        ans = ask_baidu(f'{year}年清明节是几月几号')[0]
        ans = re.findall(".*月(.)日.*", ans)
        while len(ans) == 0:
            ans = ask_baidu(f'{year}年清明节是几月几号')[0]
            ans = re.findall(".*月(.)日.*", ans)
        qingming_dict[str(year)] = ans[0]
    print(qingming_dict)  # 即qm
    """
    return datetime(year, 4, int(qm[str(year)]))


def get_solar_date(year, month, day) -> datetime:
    """
    根据农历日期获得阳历日期

    :param year:
    :param month:
    :param day:
    :return:
    """
    lunar1 = sxtwl.Lunar()
    date_solar = lunar1.getDayByLunar(year=year, month=month, day=day, isRun=False)  # isRun表示是否是闰月
    return datetime(date_solar.y, date_solar.m, date_solar.d)


def generate_stock_holiday_table(start_day=datetime(2015, 1, 1), days=7300, file='holiday.csv'):
    """
    生成股市是否开盘的csv数据表格。以下法定假日休市（2014年1月1日起施行）：
    （一）元旦，放假1天(1月1日)；
    （二）春节，放假3天(农历正月初一、初二、初三)；
    （三）清明节，放假1天(农历清明当日)；
    （四）劳动节，放假1天(5月1日)；
    （五）端午节，放假1天(农历端午当日)；
    （六）中秋节，放假1天(农历中秋当日)；
    （七）国庆节，放假3天(10月1日、2日、3日)。

    额外假期：
    2015年9月3日、9月4日，纪念抗战胜利70周年，法定休假1天调休1天，股市休市

    注：2028、2039、2047有闰五月，不知道会不会导致端午判断有问题
    下一个闰八月在2052年，暂时不会导致中秋节有问题
    下一个闰正月在2262年，与我们没有关系

    :param start_day: 从指定日期开始计算，起始日期不能早于2014年（可以等于2014），因为最新的法定放假政策从2014年开始实行
    :param days: 生成多少天的数据
    :param file: 生成的休市日期信息保存到文件file中，isOpen=1是开市，0是休市
    """
    isOpen = {}
    for i in range(days):
        day_datetime = start_day + timedelta(days=i)
        day_str = f"{day_datetime.year}/{day_datetime.month}/{day_datetime.day}"
        if day_datetime.weekday() == 5 or day_datetime.weekday() == 6:
            # 如果是星期六或者星期日，则先初始化为休市日
            isOpen[day_str] = False
        else:
            # 否则，先初始化为交易日
            isOpen[day_str] = True
    # 循环每一年
    for year in range(start_day.year, (start_day + timedelta(days=days)).year + 1):
        # 元旦节，劳动，端午，中秋，清明，均法定放假一天。这里不能算国庆节，因为国庆节放三天
        holidays = [datetime(year, 1, 1), datetime(year, 5, 1),  # datetime(year, 10, 1),
                    get_solar_date(year, 5, 5), get_solar_date(year, 8, 15),
                    get_qingming(year)]

        for select_holiday in holidays:
            weekday = select_holiday.weekday()
            # 1.如果在星期六，日，则在下一个星期一补休；
            if weekday == 5:
                holiday_datetime = [select_holiday + timedelta(days=2)]
            elif weekday == 6:
                # 例外：2015年中秋在周日，没有调休，只放假了两天
                if select_holiday == datetime(2015, 9, 27):
                    holiday_datetime = []
                else:
                    holiday_datetime = [select_holiday + timedelta(days=1)]
            # 2. 如果在星期一，五，则正常休假
            elif weekday == 0 or weekday == 4:
                holiday_datetime = [select_holiday]
            # 3. 如果在星期二，四，则分别将前一个星期六、后一个星期日调整到星期一、星期五
            elif weekday == 1:  # 因为调休的周六周日无论如何都休市，所以，这里只需要把星期一改成休市即可
                holiday_datetime = [select_holiday, select_holiday - timedelta(days=1)]
            elif weekday == 3:
                holiday_datetime = [select_holiday, select_holiday + timedelta(days=1)]
            # 4. 如果在星期三，则不调休
            # 例外：2019年劳动节是星期三，放假放的是345。
            elif weekday == 2:
                if select_holiday == datetime(2019, 5, 1):
                    holiday_datetime = [select_holiday, select_holiday + timedelta(days=1),
                                        select_holiday + timedelta(days=2)]
                else:  # 2020.1.1
                    holiday_datetime = [select_holiday]
            else:
                raise Exception("星期判断失败！")
            for day_datetime in holiday_datetime:
                day_str1 = f"{day_datetime.year}/{day_datetime.month}/{day_datetime.day}"
                isOpen[day_str1] = False
        # 国庆，将国庆所在的周从星期日到星期六放假、前后的星期六、日工作（来自百度文库，但是是错误的）
        # 根据2018、2019、2020年国庆放假规律，是10.1-10.7放假7天，多放的天数在前后两个星期内调休
        holiday_datetime = [datetime(year, 10, 1) + timedelta(days=i) for i in range(7)]
        if get_solar_date(year, 8, 15) in holiday_datetime:  # 如果中秋在国庆假期里，则10月8日再休一个天，例如2017和2020年
            holiday_datetime.append(datetime(year, 10, 8))
        for day_datetime in holiday_datetime:
            day_str2 = f"{day_datetime.year}/{day_datetime.month}/{day_datetime.day}"
            isOpen[day_str2] = False
        # 春节，百度文库的放假规律也是错误的
        # 根据2017、2018、2019年春节的放假规律，是除夕开始放假7天，2020年因为新冠疫情原因例外
        select_holiday = get_solar_date(year, 1, 1)
        if select_holiday.year == 2014:  # 2014年是从春节开始放假
            holiday_datetime = [select_holiday + timedelta(days=i) for i in range(7)]
        elif select_holiday.year >= 2015:
            holiday_datetime = [select_holiday + timedelta(days=i - 1) for i in range(7)]
        for day_datetime in holiday_datetime:
            day_str = f"{day_datetime.year}/{day_datetime.month}/{day_datetime.day}"
            isOpen[day_str] = False
    # ============================================== 例外情况 ======================================================
    # 2020年1月31日原计划正常开市，因为疫情原因休市未开，2月3日正常开市
    # 2020年5月4日和5月5日调休了两天
    if start_day.year < 2020 < (start_day + timedelta(days=days)).year + 1:
        isOpen['2020/1/31'] = False
        isOpen['2020/5/4'] = False
        isOpen['2020/5/5'] = False

    # 将holiday_dict写入csv文件
    # 将isOpen的值转为整形，方便存储
    for k, v in isOpen.items():
        if isOpen[k]:
            isOpen[k] = 1
        else:
            isOpen[k] = 0
    import pandas as pd

    df_holiday = pd.DataFrame.from_dict({"date": list(isOpen.keys()), "isOpen": list(isOpen.values())})

    df_holiday.to_csv(file, index=False)


if __name__ == '__main__':
    show_month()
    generate_stock_holiday_table()



from pycocotools.coco import COCO
import matplotlib.pyplot as plt
import cv2
import os
import numpy as np
import random
import yangke.common.fileOperate as fo
from tqdm import tqdm
from yangke.base import *
import torch


# def loadPascalVOC(dataset_folder:str):
#     """
#     加载Pascal VOC数据集，返回图像中的目标，bbox等信息
#
#     :param dataset_folder: 数据集所在的目录，如C://VOC2007
#     :return:
#     """
#     def load_single_file(xml_file: str):
#


class DataSet_YOLO:
    """
    初始化训练用的数据集，包括训练集、测试集、验证集
    """

    def __init__(self, image_folder, instance_file, img_size=416, hyp=None, augment=True, rect=True):
        """
        构建数据集

        :param image_folder: 图片文件所在的文件夹
        :param instance_file: coco数据集中的instance文件
        """
        self.img_size = img_size
        self.augment = augment
        self.rect = rect
        self.mosaic = self.augment and not self.rect  # 镶嵌图案/马赛克 (only during training)
        self.hyp = hyp

        coco = COCO(instance_file)
        """
        coco.imgs 数据集中图片信息
        coco.cats 类别信息，类别编号及类别对应的物体
        coco.anns 数据集中标注的物体位置信息，包括segmentation,bbox等，按照注释编号排列
        coco.catToImgs 数据集中每一类物体对应的图片信息（即包含该物体的所有图片的id）
        coco.dataset 数据集中所有信息，一般不用
        coco.imgToAnns 数据集中，每一张图片中的物体的位置及分类信息，按照图片id排列
        """
        self.cats = coco.cats  # 类别信息
        self.classes = len(coco.cats)  # 类别总数
        self.img_files = self.__get_images__(coco.imgs, image_folder)  # 图片的绝对路径列表
        self.images = None
        self.labels = self.__get_labels__(coco.imgToAnns)  # 标签信息，darknet中需要保证图片和labels的顺序一一对应

    def __len__(self):
        return len(self.img_files)

    def __getitem__(self, index):

        hyp = self.hyp
        if self.mosaic:
            # Load mosaic
            img, labels = self.load_mosaic(self, index)
            shapes = None

        else:
            # Load image
            img, (h0, w0), (h, w) = load_image(self, index)

            # Letterbox
            shape = self.batch_shapes[self.batch[index]] if self.rect else self.img_size  # final letterboxed shape
            img, ratio, pad = letterbox(img, shape, auto=False, scaleup=self.augment)
            shapes = (h0, w0), ((h / h0, w / w0), pad)  # for COCO mAP rescaling

            # Load labels
            labels = []
            x = self.labels[index]  # x对应class, x_center, y_center, width, height，其中x,y,w,h是归一化后的数据
            if x is not None and x.size > 0:
                # 数据集的labels文件中坐标是归一化的xywh，即x_center, y_center, width, height
                # 这里将归一化的wywh数据转换为像素为单位的 xyxy数据(矩形左上角xy和右下角xy)
                # 这里得到的像素坐标不是整数，直接取整即可，对实际目标边框的影响可以忽略
                labels = x.copy()
                labels[:, 1] = ratio[0] * w * (x[:, 1] - x[:, 3] / 2) + pad[0]  # pad width
                labels[:, 2] = ratio[1] * h * (x[:, 2] - x[:, 4] / 2) + pad[1]  # pad height
                labels[:, 3] = ratio[0] * w * (x[:, 1] + x[:, 3] / 2) + pad[0]
                labels[:, 4] = ratio[1] * h * (x[:, 2] + x[:, 4] / 2) + pad[1]

        if self.augment:
            # Augment imagespace
            if not self.mosaic:
                img, labels = random_affine(img, labels,
                                            degrees=hyp['degrees'],
                                            translate=hyp['translate'],
                                            scale=hyp['scale'],
                                            shear=hyp['shear'])

            # Augment colorspace
            augment_hsv(img, hgain=hyp['hsv_h'], sgain=hyp['hsv_s'], vgain=hyp['hsv_v'])

            # Apply cutouts
            # if random.random() < 0.9:
            #     labels = cutout(img, labels)

        nL = len(labels)  # number of labels
        if nL:
            # convert xyxy to xywh，仍然是像素为单位
            labels[:, 1:5] = xyxy2xywh(labels[:, 1:5])

            # Normalize coordinates 0 - 1，归一化到0-1
            labels[:, [2, 4]] /= img.shape[0]  # height
            labels[:, [1, 3]] /= img.shape[1]  # width

        if self.augment:
            # random left-right flip
            lr_flip = True
            if lr_flip and random.random() < 0.5:
                img = np.fliplr(img)
                if nL:
                    labels[:, 1] = 1 - labels[:, 1]

            # random up-down flip
            ud_flip = False
            if ud_flip and random.random() < 0.5:
                img = np.flipud(img)
                if nL:
                    labels[:, 2] = 1 - labels[:, 2]

        labels_out = torch.zeros((nL, 6))  # n行6列的Tensor，n对应图片中的目标数
        if nL:
            labels_out[:, 1:] = torch.from_numpy(labels)

        # Convert
        img = img[:, :, ::-1].transpose(2, 0, 1)  # BGR to RGB, to 3x416x416
        img = np.ascontiguousarray(img)
        # imgs, targets, paths, _
        return torch.from_numpy(img), labels_out, self.img_files[index], shapes

    @staticmethod
    def collate_fn(batch):
        """
        取数据的函数

        :param batch: batch是一批训练数据，batch[i]对应一张图片和图片标签，batch[i]是self.__getitem__()方法返回的结果
        :return:
        """
        img, label, path, shapes = zip(*batch)  # transposed
        # 假设batch_size=16，这里依次将16张图片的label[0]设置为i，这样当前图片中全部n个目标的label[0]都指到i
        for i, l in enumerate(label):
            l[:, 0] = i  # add target image index for build_targets()图片中有多少目标，l就有多少行，l的第一列都指到当前图片
        return torch.stack(img, 0), torch.cat(label, 0), path, shapes

    def __get_images__(self, imgs: dict, folders):
        """
        从COCO数据集的imgs中抽取出所有图片文件路径，保存到列表中返回

        :param imgs: COCO数据集的imgs对象，是一个字典
        :param folders: COCO数据集图片存储路径
        :return: 所有图片的绝对路径的列表
        """
        images = []
        if not isinstance(folders, list):
            folders = [folders]
        for k, v in imgs.items():
            file_name = v.get('file_name')
            for folder in folders:
                file_name = os.path.join(folder, file_name)
                if os.path.exists(file_name):
                    images.append(file_name)
                    break
        return images

    def __get_labels__(self, imgToAnns):
        """
        根据instance文件收集边框信息

        :param coco:
        :return:
        """
        labels = []
        for k, v in imgToAnns.items():
            # k是图片id
            # v是图片中的待检测目标信息，因为一张图片中包含多个目标，因此v是一个列表
            bboxs = []
            for obj in v:
                # obj是一个字典
                img_id = obj.get('image_id')
                cat_id = obj.get('category_id')
                bbox = obj.get('bbox')
                bboxs.append([img_id, cat_id].extend(bbox))
            labels.append(bboxs)
        return labels

    def load_image(self, index):
        """
        加载指定索引的图片，如果图片大于512，则缩小到512，如果图片小于512且开启了图像增强，则保证长宽比不变的同时将最大边放大到512

        :param self: 数据集，Dataset子类
        :param index: 图片索引
        :return: 缩放后的图片，（原图高、宽），缩放后图片的宽、高
        """
        # loads 1 image from dataset, returns img, original hw, resized hw

        img_path = self.img_files[index]
        img = cv2.imread(img_path)  # BGR
        assert img is not None, 'Image Not Found ' + img_path
        h0, w0 = img.shape[:2]  # orig hw
        r = self.img_size / max(h0, w0)  # resize image to img_size
        if r < 1 or (self.augment and r != 1):  # always resize down, only resize up if training with augmentation
            interp = cv2.INTER_AREA if r < 1 and not self.augment else cv2.INTER_LINEAR
            img = cv2.resize(img, (int(w0 * r), int(h0 * r)), interpolation=interp)
        return img, (h0, w0), img.shape[:2]  # img, hw_original, hw_resized


def load_mosaic(self, index):
    """
    将4张图片拼接成一张图片

    :param self:
    :param index:
    :return:
    """
    # loads images in a mosaic, https://www.cnblogs.com/wujianming-110117/p/12806502.html

    labels4 = []
    s = self.img_size
    xc, yc = [int(random.uniform(s * 0.5, s * 1.5)) for _ in range(2)]  # mosaic center x, y
    indices = [index] + [random.randint(0, len(self.labels) - 1) for _ in
                         range(3)]  # 3 additional image indices，随机3张附加图片
    for i, index in enumerate(indices):
        # Load image
        img, _, (h, w) = load_image(self, index)

        # place img in img4
        if i == 0:  # top left
            img4 = np.full((s * 2, s * 2, img.shape[2]), 114, dtype=np.uint8)  # base image with 4 tiles
            x1a, y1a, x2a, y2a = max(xc - w, 0), max(yc - h, 0), xc, yc  # xmin, ymin, xmax, ymax (large image)
            x1b, y1b, x2b, y2b = w - (x2a - x1a), h - (y2a - y1a), w, h  # xmin, ymin, xmax, ymax (small image)
        elif i == 1:  # top right
            x1a, y1a, x2a, y2a = xc, max(yc - h, 0), min(xc + w, s * 2), yc
            x1b, y1b, x2b, y2b = 0, h - (y2a - y1a), min(w, x2a - x1a), h
        elif i == 2:  # bottom left
            x1a, y1a, x2a, y2a = max(xc - w, 0), yc, xc, min(s * 2, yc + h)
            x1b, y1b, x2b, y2b = w - (x2a - x1a), 0, max(xc, w), min(y2a - y1a, h)
        elif i == 3:  # bottom right
            x1a, y1a, x2a, y2a = xc, yc, min(xc + w, s * 2), min(s * 2, yc + h)
            x1b, y1b, x2b, y2b = 0, 0, min(w, x2a - x1a), min(y2a - y1a, h)

        img4[y1a:y2a, x1a:x2a] = img[y1b:y2b, x1b:x2b]  # img4[ymin:ymax, xmin:xmax]
        padw = x1a - x1b
        padh = y1a - y1b

        # Load labels
        label_path = self.label_files[index]
        if os.path.isfile(label_path):
            x = self.labels[index]
            if x is None:  # labels not preloaded
                with open(label_path, 'r') as f:
                    x = np.array([x.split() for x in f.read().splitlines()], dtype=np.float32)

            if x.size > 0:
                # Normalized xywh to pixel xyxy format
                labels = x.copy()
                labels[:, 1] = w * (x[:, 1] - x[:, 3] / 2) + padw
                labels[:, 2] = h * (x[:, 2] - x[:, 4] / 2) + padh
                labels[:, 3] = w * (x[:, 1] + x[:, 3] / 2) + padw
                labels[:, 4] = h * (x[:, 2] + x[:, 4] / 2) + padh
            else:
                labels = np.zeros((0, 5), dtype=np.float32)
            labels4.append(labels)

    # Concat/clip labels
    if len(labels4):
        labels4 = np.concatenate(labels4, 0)
        # np.clip(labels4[:, 1:] - s / 2, 0, s, out=labels4[:, 1:])  # use with center crop
        np.clip(labels4[:, 1:], 0, 2 * s, out=labels4[:, 1:])  # use with random_affine

    # Augment
    # img4 = img4[s // 2: int(s * 1.5), s // 2:int(s * 1.5)]  # center crop (WARNING, requires box pruning)
    img4, labels4 = random_affine(img4, labels4,
                                  degrees=self.hyp['degrees'] * 1,
                                  translate=self.hyp['translate'] * 1,
                                  scale=self.hyp['scale'] * 1,
                                  shear=self.hyp['shear'] * 1,
                                  border=-s // 2)  # border to remove

    return img4, labels4


def loadCOCO(coco_folder: str):
    pass


def coco2darknet(train_image_folder=None, test_image_folder=None, val_image_folder=None,
                 instance_annotations_files: list = None, dest_folder: str = None):
    """
    根据COCO数据集生成Darknet数据集

    darknet数据集包括
    图片路径 ： image_folder
    图片中物体标签： label_folder

    训练用图片集： train_image.txt
    测试用图片集： test_image.txt

    :param val_image_folder:
    :param test_image_folder:
    :param train_image_folder:
    :param instance_annotations_files: coco数据集根目录，是
    :param dest_folder: 生成的darknet数据集保存目录
    :return:
    """
    if isinstance(instance_annotations_files, str):
        instance_annotations_files = [instance_annotations_files]

    dest_label_train_folder = os.path.join(dest_folder, 'labels/train')
    dest_label_test_folder = os.path.join(dest_folder, 'labels/test')
    dest_label_val_folder = os.path.join(dest_folder, 'labels/val')
    dest_image_folder = os.path.join(dest_folder, 'images')
    os.makedirs(dest_label_train_folder, exist_ok=True)
    os.makedirs(dest_image_folder, exist_ok=True)
    os.makedirs(dest_label_test_folder, exist_ok=True)
    os.makedirs(dest_label_val_folder, exist_ok=True)

    # 一般COCO数据集目标检测有两个注释文件，分别为instances_train2017.json和instances_val2017.json
    for anns_file in instance_annotations_files:
        assert os.path.exists(anns_file), f"annotations文件不存在，文件路径：{anns_file}"
        if 'train' in instance_annotations_files:
            dataset_type = 'train'
            current_image_folder = train_image_folder
            current_dest_label_folder = dest_label_train_folder
        elif 'test' in instance_annotations_files:
            dataset_type = 'test'
            current_image_folder = test_image_folder
            current_dest_label_folder = dest_label_test_folder
        else:
            dataset_type = 'val'
            current_image_folder = val_image_folder
            current_dest_label_folder = dest_label_val_folder
        coco = COCO(anns_file)
        txt_imgpath = f'{dataset_type}_image.txt'
        txt_labelpath = f'{dataset_type}_label.txt'
        images = set()

        for k, v in tqdm(coco.anns.items()):  # 遍历所有的annotations
            anns_id = k
            seg_coor_list: list = v.get('segmentation')  # 二维列表，一个物体可能有多个多边形组成
            img_id: int = v.get('image_id')
            bbox: list = v.get('bbox')
            category_id: int = v.get('category_id')
            category = coco.cats.get(category_id)  # 根据category_id拿到具体的分类
            # 根据img_id拿到对应的image对象
            image = coco.imgs.get(img_id)
            # 获取图片的绝对路径
            img_file = os.path.join(current_image_folder, image.get('file_name'))
            # 检测本地图片是否存在
            if not os.path.exists(img_file):
                temp_ = os.path.join(train_image_folder, os.path.basename(img_file))
                if os.path.exists(temp_):
                    print(f'{img_file} 所在文件夹错误，文件位于{temp_}')
                temp_ = os.path.join(val_image_folder, os.path.basename(img_file))
                if os.path.exists(temp_):
                    print(f'{img_file} 所属文件夹错误，文件位于{temp_}')
                print(f'{img_file} 不存在')
                continue
            images.add(img_file)  # 将图片添加到图片集中，集合会自动去重
            img_height = image.get('height')
            img_width = image.get('width')
            img_url = image.get('coco_url')

            # 将bbox归一化到0~1
            bbox[0] /= img_width
            bbox[2] /= img_width
            bbox[1] /= img_height
            bbox[3] /= img_height

            dest_label_filename = os.path.splitext(image.get('file_name'))[0] + '.txt'
            dest_label_file = os.path.join(current_dest_label_folder, dest_label_filename)

            with open(dest_label_file, 'a+') as f:  # 写labels文件
                f.write(f'{category_id} {bbox[0]:8.6f} {bbox[1]:8.6f} {bbox[2]:8.6f} {bbox[3]:8.6f}\n')
        fo.write_lines(txt_imgpath, list(images))  # 写训练图片 集合

# coco2darknet(train_image_folder=r'D:\Users\pycharm\face\data\train2017',
#              test_image_folder=r'D:\Users\pycharm\face\data\test2017',
#              val_image_folder=r'D:\Users\pycharm\face\data\val2017',
#              instance_annotations_files=[  # r'D:\Users\pycharm\face\data\annotations\instances_train2017.json',
#                  r'D:\Users\pycharm\face\data\annotations\instances_val2017.json'],
#              dest_folder=r'D:\Users\pycharm\face\data\dark')

# data = DataSet_YOLO(image_folder=r'D:\Users\pycharm\face\data\train2017',
#                     instance_file=r'D:\Users\pycharm\face\data\annotations\instances_train2017.json')



from enum import Enum, unique
from yangke.base import get_key_value, is_number
from yangke.common.config import logger

from enum import Enum, unique
from yangke.base import get_key_value, is_number
from yangke.common.config import logger


@unique
@get_key_value
class TagsRead(Enum):
    # 优化目标 = "obj"  # 0表示煤耗，1表示收益
    # price_coal = "DCS.PRICE_COAL"  # 元/t
    # price_power = "DCS.PRICE_POWER"  # 元/kWh

    power3 = "N3DCS.MW"
    power4 = "N4DCS.MW"
    # bpl_a_3 = "N3DCS.30MAG10CP101"  # 低压凝汽器背压a
    # bpl_b_3 = "N3DCS.30MAG10CP102"  # 低压凝汽器背压b
    # bpl_c_3 = "N3DCS.30MAG10CP103"  # 低压凝汽器背压c
    # bph_a_3 = "N3DCS.30MAG20CP101"  # 高压凝汽器背压a
    # bph_b_3 = "N3DCS.30MAG20CP102"  # 高压凝汽器背压b
    # bph_c_3 = "N3DCS.30MAG20CP103"  # 高压凝汽器背压c
    #
    # bpl_a_4 = "N4DCS.40MAG10CP101"  # 低压凝汽器背压a
    # bpl_b_4 = "N4DCS.40MAG10CP102"  # 低压凝汽器背压b
    # bpl_c_4 = "N4DCS.40MAG10CP103"  # 低压凝汽器背压c
    # bph_a_4 = "N4DCS.40MAG20CP101"  # 高压凝汽器背压a
    # bph_b_4 = "N4DCS.40MAG20CP102"  # 高压凝汽器背压b
    # bph_c_4 = "N4DCS.40MAG20CP103"  # 高压凝汽器背压c
    #
    # 凝泵出口凝结水流量3 = "N3DCS.CNDFLW2"  #
    # 凝泵出口凝结水流量4 = "N4DCS.CNDFLW2"
    # 热井出水温度3 = "N3DCS.30LCA11CT101"
    # 热井出水温度4 = "N4DCS.40LCA11CT101"
    # 凝汽器运行补水流量3 = "N3DCS.30LCP10CF102_C"
    # 凝汽器补水温度3 = "N3DCS.30LCA20CT101"  # 压力取2.28MPa
    # 凝汽器运行补水流量4 = "N4DCS.40LCP10CF102_C"  # 98.25
    # 凝汽器补水温度4 = "N4DCS.40LCA20CT101"  # 24.68
    #
    # 大气压力 = "N3DCS.30HLA00CP101"  # 3号炉风烟系统画面
    # 环境温度 = "N3DCS.30HLA00CT101"  # 3号炉风烟系统画面
    # 环境湿度 = "ldyh.gtc_11mbl01cm101"
    #
    # 凝汽器1进口A循环水温度1_3 = "N3DCS.30PAB10CT101"  # 凝汽器1为低压凝汽器，循环水先进入低压凝汽器在进入高压凝汽器
    # 凝汽器1进口A循环水温度2_3 = "N3DCS.30PAB10CT102"
    # 凝汽器1进口B循环水温度1_3 = "N3DCS.30PAB20CT101"
    # 凝汽器1进口B循环水温度2_3 = "N3DCS.30PAB20CT102"
    #
    # 凝汽器2出口A循环水温度1_3 = "N3DCS.30PAB10CT103"  # 凝汽器2为高压凝汽器
    # 凝汽器2出口A循环水温度2_3 = "N3DCS.30PAB10CT104"
    # 凝汽器2出口B循环水温度1_3 = "N3DCS.30PAB20CT103"
    # 凝汽器2出口B循环水温度2_3 = "N3DCS.30PAB20CT104"
    #
    # 凝汽器1进口A循环水温度1_4 = "N4DCS.40PAB10CT101"
    # 凝汽器1进口A循环水温度2_4 = "N4DCS.40PAB10CT102"
    # 凝汽器1进口B循环水温度1_4 = "N4DCS.40PAB20CT101"
    # 凝汽器1进口B循环水温度2_4 = "N4DCS.40PAB20CT102"
    #
    # 凝汽器2出口A循环水温度1_4 = "N4DCS.30PAB10CT103"
    # 凝汽器2出口A循环水温度2_4 = "N4DCS.30PAB10CT104"
    # 凝汽器2出口B循环水温度1_4 = "N4DCS.30PAB20CT103"
    # 凝汽器2出口B循环水温度2_4 = "N4DCS.30PAB20CT104"
    #
    # 循泵A电流3 = "N3DCS.30PAC01AP001XQ01"
    # 循泵B电流3 = "N3DCS.30PAC01AP002XQ01"
    # 循泵C电流3 = "N3DCS.30PAC01AP003XQ01"
    # 循泵A电流4 = "N4DCS.40PAC01AP001XQ01"
    # 循泵B电流4 = "N4DCS.40PAC01AP002XQ01"
    # 循泵C电流4 = "N4DCS.40PAC01AP003XQ01"
    #
    # 循泵A转速3 = "N3DCS.30PAC01CS111"  # 典型转速372
    # 循泵B转速3 = "N3DCS.30PAC01CS112"  # 高速：转速>350:电流>150;典型转速372；低速:典型转速330
    # 循泵C转速3 = "N3DCS.30PAC01CS113"
    # 循泵A转速4 = "N4DCS.40PAC01CS111"
    # 循泵B转速4 = "N4DCS.40PAC01CS112"
    # 循泵C转速4 = "N4DCS.40PAC01CS113"

    # 入炉煤量 = ""


@unique
@get_key_value
class TagsReadDCS(Enum):
    # 优化目标 = "obj"  # 0表示煤耗，1表示收益
    # price_coal = "DCS.PRICE_COAL"  # 元/t
    # price_power = "DCS.PRICE_POWER"  # 元/kWh

    power3 = "CF159F30MKA01CE001"
    power4 = "CF159FD40XKA01CE903_XQ01"  # CF159FD40CBA25CBA25R101
    bpl_a_3 = "CF159FCNSRVCM"  # 低压凝汽器背压a
    # bpl_b_3 = "N3DCS.30MAG10CP102"  # 低压凝汽器背压b
    # bpl_c_3 = "N3DCS.30MAG10CP103"  # 低压凝汽器背压c
    # bph_a_3 = "N3DCS.30MAG20CP101"  # 高压凝汽器背压a
    # bph_b_3 = "N3DCS.30MAG20CP102"  # 高压凝汽器背压b
    # bph_c_3 = "N3DCS.30MAG20CP103"  # 高压凝汽器背压c
    #
    # bpl_a_4 = "N4DCS.40MAG10CP101"  # 低压凝汽器背压a
    # bpl_b_4 = "N4DCS.40MAG10CP102"  # 低压凝汽器背压b
    # bpl_c_4 = "N4DCS.40MAG10CP103"  # 低压凝汽器背压c
    # bph_a_4 = "N4DCS.40MAG20CP101"  # 高压凝汽器背压a
    # bph_b_4 = "N4DCS.40MAG20CP102"  # 高压凝汽器背压b
    # bph_c_4 = "N4DCS.40MAG20CP103"  # 高压凝汽器背压c
    #
    # 凝泵出口凝结水流量3 = "N3DCS.CNDFLW2"  #
    # 凝泵出口凝结水流量4 = "N4DCS.CNDFLW2"
    # 热井出水温度3 = "N3DCS.30LCA11CT101"
    # 热井出水温度4 = "N4DCS.40LCA11CT101"
    # 凝汽器运行补水流量3 = "N3DCS.30LCP10CF102_C"
    # 凝汽器补水温度3 = "N3DCS.30LCA20CT101"  # 压力取2.28MPa
    # 凝汽器运行补水流量4 = "N4DCS.40LCP10CF102_C"  # 98.25
    # 凝汽器补水温度4 = "N4DCS.40LCA20CT101"  # 24.68
    #
    # 大气压力 = "N3DCS.30HLA00CP101"  # 3号炉风烟系统画面
    # 环境温度 = "N3DCS.30HLA00CT101"  # 3号炉风烟系统画面
    # 环境湿度 = "ldyh.gtc_11mbl01cm101"
    #
    # 凝汽器1进口A循环水温度1_3 = "N3DCS.30PAB10CT101"  # 凝汽器1为低压凝汽器，循环水先进入低压凝汽器在进入高压凝汽器
    # 凝汽器1进口A循环水温度2_3 = "N3DCS.30PAB10CT102"
    # 凝汽器1进口B循环水温度1_3 = "N3DCS.30PAB20CT101"
    # 凝汽器1进口B循环水温度2_3 = "N3DCS.30PAB20CT102"
    #
    # 凝汽器2出口A循环水温度1_3 = "N3DCS.30PAB10CT103"  # 凝汽器2为高压凝汽器
    # 凝汽器2出口A循环水温度2_3 = "N3DCS.30PAB10CT104"
    # 凝汽器2出口B循环水温度1_3 = "N3DCS.30PAB20CT103"
    # 凝汽器2出口B循环水温度2_3 = "N3DCS.30PAB20CT104"
    #
    # 凝汽器1进口A循环水温度1_4 = "N4DCS.40PAB10CT101"
    # 凝汽器1进口A循环水温度2_4 = "N4DCS.40PAB10CT102"
    # 凝汽器1进口B循环水温度1_4 = "N4DCS.40PAB20CT101"
    # 凝汽器1进口B循环水温度2_4 = "N4DCS.40PAB20CT102"
    #
    # 凝汽器2出口A循环水温度1_4 = "N4DCS.30PAB10CT103"
    # 凝汽器2出口A循环水温度2_4 = "N4DCS.30PAB10CT104"
    # 凝汽器2出口B循环水温度1_4 = "N4DCS.30PAB20CT103"
    # 凝汽器2出口B循环水温度2_4 = "N4DCS.30PAB20CT104"
    #
    # 循泵A电流3 = "N3DCS.30PAC01AP001XQ01"
    # 循泵B电流3 = "N3DCS.30PAC01AP002XQ01"
    # 循泵C电流3 = "N3DCS.30PAC01AP003XQ01"
    # 循泵A电流4 = "N4DCS.40PAC01AP001XQ01"
    # 循泵B电流4 = "N4DCS.40PAC01AP002XQ01"
    # 循泵C电流4 = "N4DCS.40PAC01AP003XQ01"
    #
    # 循泵A转速3 = "N3DCS.30PAC01CS111"  # 典型转速372
    # 循泵B转速3 = "N3DCS.30PAC01CS112"  # 高速：转速>350:电流>150;典型转速372；低速:典型转速330
    # 循泵C转速3 = "N3DCS.30PAC01CS113"
    # 循泵A转速4 = "N4DCS.40PAC01CS111"
    # 循泵B转速4 = "N4DCS.40PAC01CS112"
    # 循泵C转速4 = "N4DCS.40PAC01CS113"

    # 入炉煤量 = ""


@unique
@get_key_value
class TagsWrite(Enum):
    循泵A启停3 = "N3DCS.PUMP_A_OPT_3"
    循泵B启停3 = "N3DCS.PUMP_B_OPT_3"
    循泵C启停3 = "N3DCS.PUMP_C_OPT_3"
    循泵A启停4 = "N4DCS.PUMP_A_OPT_4"
    循泵B启停4 = "N4DCS.PUMP_B_OPT_4"
    循泵C启停4 = "N4DCS.PUMP_C_OPT_4"

    凝汽器热负荷3 = "N3DCS.HEAT_LOAD_COND_3"
    凝汽器热负荷4 = "N3DCS.HEAT_LOAD_COND_4"
    凝汽器端差3 = 'N3DCS.DC_CONDENSER_3'
    凝汽器端差4 = 'N3DCS.DC_CONDENSER_4'
    凝汽器过冷度3 = 'N3DCS.SUPERCOOL_CONDENSER_3'
    凝汽器过冷度4 = 'N3DCS.SUPERCOOL_CONDENSER_4'
    循环水流量3 = 'N3DCS.FLOW_CYC_3'
    循环水流量4 = 'N3DCS.FLOW_CYC_4'

    冷端耗功3 = 'N3DCS.POWER_COLD_3'
    冷端耗功优化值3 = 'N3DCS.POWER_COLD_OPT_3'
    发电功率优化值3 = 'N3DCS.POWER_OPT_3'
    净功率3 = 'N3DCS.POWER_NET_3'
    净功率优化值3 = 'N3DCS.POWER_NET_OPT_3'
    煤耗3 = 'N3DCS.COAL_CON_3'
    煤耗优化值3 = 'N3DCS.COAL_CON_OPT_3'
    发电收益3 = 'N3DCS.BENEFIT_GEN_3'
    发电收益优化值3 = 'N3DCS.BENEFIT_GEN_OPT_3'

    冷端耗功4 = 'N3DCS.POWER_COLD_4'
    冷端耗功优化值4 = 'N3DCS.POWER_COLD_OPT_4'
    发电功率优化值4 = 'N3DCS.POWER_OPT_4'
    净功率4 = 'N3DCS.POWER_NET_4'
    净功率优化值4 = 'N3DCS.POWER_NET_OPT_4'
    煤耗4 = 'N3DCS.COAL_CON_4'
    煤耗优化值4 = 'N3DCS.COAL_CON_OPT_4'
    发电收益4 = 'N3DCS.BENEFIT_GEN_4'
    发电收益优化值4 = 'N3DCS.BENEFIT_GEN_OPT_4'



from pylatex import Document, Section, Subsection, Command
from pylatex.utils import italic, NoEscape


def fill_document(doc):
    with doc.create(Section('A section')):
        doc.append('Some regular text and some')

        with doc.create(Subsection('A subsection')):
            doc.append('Also some crazy characters:$&#{}')


if __name__ == "__main__":
    doc = Document('basic')
    fill_document(doc)
    doc.generate_pdf(clean_tex=False)
    doc.generate_tex()




def set_LSTM_x_y(data_local):
    data_local = data_local.tolist()
    size = len(data_local)
    x = []  # x包含data_local个数据
    y = []
    local_kind = ["跌", "涨幅位于区间[0, 1%]", "涨幅位于区间[1%, 2%]",
                  "涨幅位于区间[2%, 3%]", "涨幅位于区间[3%, 4%]",
                  "涨幅位于区间[4%, 5%]", "涨幅位于区间[5%, 6%]", "涨幅位于区间[6%, 7%]",
                  "涨幅位于区间[7%, 8%]", "涨幅位于区间[8%, 10%"]
    for i in range(size - 60):
        x_temp = []
        for j in range(60, 0, -1):
            x_temp.append(data_local[i + 60 - j])
        x.append(x_temp)
        inc = data_local[i + 60][0] - data_local[i + 59][0]
        y_temp = inc / data_local[i + 59][0]
        if y_temp < 0:
            y.append(0)
        elif y_temp < 0.01:
            y.append(1)
        elif y_temp < 0.02:
            y.append(2)
        elif y_temp < 0.03:
            y.append(3)
        elif y_temp < 0.04:
            y.append(4)
        elif y_temp < 0.05:
            y.append(5)
        elif y_temp < 0.06:
            y.append(6)
        elif y_temp < 0.07:
            y.append(7)
        elif y_temp < 0.08:
            y.append(8)
        else:
            y.append(9)
    # x比y多一个数据
    x_temp = []
    for j in range(60, 0, -1):
        x_temp.append(data_local[size - j])

    x.append(x_temp)
    return x, y, local_kind


def normalization(x):
    x = np.array(x)
    mean = np.mean(x, axis=0)
    std = np.std_range(x, axis=0)
    normalized_x = (x - mean) / std
    return normalized_x


def prediction_by_1MACD(df):
    """
    MACD_DIFF、MACD_DEA、MACD;中长期指标

    :return label: 0-无预测意见，1-买入信号，-1-卖出信号
    """
    label = 0
    diff = np.array(df['MACD_DIFF'])
    dea = np.array(df['MACD_DEA'])
    macd = np.array(df['MACD'])
    if diff[-1] > 0 & dea[-1] > 0:
        if (diff[-1] - dea[-1]) > (diff[-2] - dea[-2]) > (diff[-3] - dea[-3]) > (diff[-4] - dea[-4]):
            label = 1 + label
    if (diff[-1] - dea[-1]) < (diff[-2] - dea[-2]) < (diff[-3] - dea[-3]):
        label = label - 1
    return label


def prediction_by_2DMI(df):
    """
    中长期指标
    :param df:
    :return:
    """
    pass


def prediction_by_3DMA(df):
    """
    中短期指标
    :param df:
    :return:
    """
    pass


def prediction_by_4TRIX(df):
    """
    长期指标
    :param df:
    :return:
    """
    TRIX = df['TRIX']
    MATRIX = df['MATRIX']
    result = 0
    if (TRIX[-1] - MATRIX[-1]) > 0 > (TRIX[-2] - MATRIX[-2]):  # 认为TRIX线向上突破TRMA线 金叉
        result = 1
    elif (TRIX[-1] - MATRIX[-1]) < 0 < (TRIX[-2] - MATRIX[-2]):  # 认为TRIX线在高位向下突破TRMA线 死叉
        result = -1
    return result


def prediction_by_5BRAR(df):
    pass


def prediction_by_6VR(df):
    """
    中期指标
    :param df:
    :return:
    """
    pass


def prediction_by_7OBV(df):
    pass


def prediction_by_8EMV(df):
    pass


def prediction_by_9WVAD(df):
    """
    长期指标
    WVAD、ADX、PDI、MDI
    :param df:
    :return:
    """
    pass


def prediction_by_10RSI(df):
    pass


def prediction_by_11WR(df):
    """
    短期指标
    :param df:
    :return:
    """
    pass


def prediction_by_12KDJ(df):
    """
    中短期指标
    :param df:
    :return:
    """
    pass


def prediction_by_13CCI(df):
    """
    短期暴涨暴跌行情研判指标
    :param df:
    :return:
    """
    pass


def prediction_by_14ROC(df):
    """
    短期指标
    :param df:
    :return:
    """
    pass


def prediction_by_15BOLL(df):
    """
    中长期指标
    :param df:
    :return:
    """
    pass


def prediction_by_16BIAS(df):
    pass


def prediction_by_17PSY(df):
    """
    短期指标
    :param df:
    :return:
    """
    pass
